import ast
import copy
import os
import time
import datetime
import logging
import queue
import sys
import yaml
import pkg_resources
from PyQt5 import QtWidgets, QtCore, QtGui
import inspect
import threading
import multiprocessing
import socket
import argparse
import importlib.util
import glob
import pathlib
import signal
import uuid
import random
import re
import pydantic
import typing
from pyqtconsole.console import PythonConsole
from pyqtconsole.highlighter import format
import platform
# Import redvypr specific stuff
import redvypr.data_packets as data_packets
import redvypr.redvypr_address as redvypr_address
import redvypr.packet_statistic as redvypr_packet_statistic
from redvypr.version import version
import redvypr.files as files
from redvypr.device import RedvyprDeviceConfig, RedvyprDeviceBaseConfig, RedvyprDevice, RedvyprDeviceScan, RedvyprDeviceParameter, queuesize
import redvypr.devices as redvyprdevices
import faulthandler
faulthandler.enable()

# Collect all logger
logger_all = [data_packets.logger, redvypr_address.logger, redvypr_packet_statistic.logger]

# Platform information str
__platform__ = "redvypr (REaltime Data Vi(Y)ewer and PRocessor (in Python))\n"
__platform__ += "\n\n"
__platform__ += "Version: {:s}\n".format(str(version))
__platform__ += "Python: {:s}\n".format(sys.version)
__platform__ += "Platform system: {:s}\n".format(platform.system())
__platform__ += "Platform release: {:s}\n".format(platform.release())
__platform__ += "Platform version: {:s}\n".format(platform.version())

_logo_file = files.logo_file
_icon_file = files.icon_file

logging.basicConfig(stream=sys.stderr)
logger = logging.getLogger('redvypr')
logger.setLevel(logging.INFO)


metadata_address = '/d:/p:/i:metadata/k:_redvypr_command'

# Pydantic

class RedvyprMetadata(pydantic.BaseModel):
    model_config = pydantic.ConfigDict(extra="allow")
    description: str = pydantic.Field(default='')
    lon: float = pydantic.Field(default=-9999)
    lat: float = pydantic.Field(default=-9999)
class RedvyprConfig(pydantic.BaseModel):
    model_config = pydantic.ConfigDict(extra="allow")
    hostname: typing.Optional[str] = pydantic.Field(default=None)
    metadata: typing.Optional[RedvyprMetadata] = pydantic.Field(default=None)
    #devices: list = pydantic.Field(default=[])
    #devices: typing.List[RedvyprDeviceConfig] = pydantic.Field(default=[])
    devices: typing.List[typing.Annotated[typing.Union[RedvyprDeviceConfig], pydantic.Field(discriminator='config_type')]] = pydantic.Field(default=[])
    devicepaths: list = pydantic.Field(default=[])
    loglevel: typing.Literal['INFO','DEBUG','WARNING'] = pydantic.Field(default='INFO')
    gui_home_icon: str = 'redvypr'



# https://stackoverflow.com/questions/166506/finding-local-ip-addresses-using-pythons-stdlib
def get_ip():
    s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    try:
        # doesn't even have to be reachable
        s.connect(('10.255.255.255', 1))
        IP = s.getsockname()[0]
    except Exception:
        IP = '127.0.0.1'
    finally:
        s.close()
    return IP


# The hostinfo, to distinguish between different redvypr instances
# redvyprid = str(uuid.uuid1()) # Old

hostinfo_blank = {'hostname':None, 'tstart': 0,'addr':None, 'uuid':None}

def create_hostinfo(hostname='redvypr'):
    funcname = __name__ + '.create_hostinfo()'
    logger.debug(funcname)
    randstr = '{:03d}'.format(random.randrange(2 ** 8))
    #redvyprid = datetime.datetime.now().strftime('%Y%m%d%H%M%S.%f-') + str(uuid.getnode()) + '-' + randstr
    redvyprid = str(uuid.getnode()) + '-' + randstr
    hostinfo = {'hostname': hostname, 'tstart': time.time(), 'addr': get_ip(), 'uuid': redvyprid}
    return hostinfo

def distribute_data(devices, hostinfo, deviceinfo_all, infoqueue, redvyprqueue, dt=0.01):
    """ The heart of redvypr, this functions distributes the queue data onto the subqueues.
    """
    funcname = __name__ + '.distribute_data()'
    datastreams_all = {}
    datastreams_all_old = {}
    dt_info = 5.0  # The time interval information will be sent
    dt_avg = 0  # Averaging of the distribution time needed
    navg = 0
    packets_processed = 0 # For statistics, count packets
    packet_counter = 0 # Global counter of packets received by the redvypr instance
    tinfo = time.time()
    tstop = time.time()
    thread_start = time.time()
    dt_sleep = dt

    while True:
        try:
            time.sleep(dt_sleep)
            tstart = time.time()
            FLAG_device_status_changed = False
            devices_changed = []
            devices_removed = []
            # Read data from the main thread
            try:
                redvyprdata = redvyprqueue.get(block=False) # Data from the main thread
            except Exception as e:
                redvyprdata = None
            # Process data from the main thread
            if(redvyprdata is not None):
                if(redvyprdata['type'] == 'device_removed'):
                    logger.debug(funcname + 'Device removed {}'.format(redvyprdata))
                    FLAG_device_status_changed = True
                    devices_removed.append(redvyprdata['device'])
                    devinfo_rem = deviceinfo_all['device_redvypr'].pop(redvyprdata['device'])
                    devinfo_send = {'type':'deviceinfo_all', 'deviceinfo_all': copy.deepcopy(deviceinfo_all), 'devices_changed': list(set(devices_changed)),
                    'devices_removed': devices_removed,'change':'devrem','device_changed':redvyprdata['device']}
                    infoqueue.put_nowait(devinfo_send)
                    # Send a deviceinfo update with the changed metadata
                    compacket = data_packets.commandpacket('info', host=hostinfo, devicename='', packetid='device_removed',
                                                           publisher='')
                    compacket['deviceinfo_all'] = copy.deepcopy(deviceinfo_all)
                    compacket['devices_removed'] = devices_removed
                    redvypr_packet_statistic.treat_datadict(compacket, '', hostinfo, 0, tread,
                                                            'distribute_data')
                    data_packets_fan_out.append(compacket)

            for devicedict in devices:
                device = devicedict['device']
                data_all = []
                tread = time.time()
                # Read all packets in a bunch
                while True:
                    try:
                        data = device.dataqueue.get(block=False)
                        if(type(data) is not dict): # If data is not a dictionary, convert it to one
                            data = {'data':data}

                        devicedict['statistics']['packets_published'] += 1 # The total number of packets published by the device
                        packets_processed += 1 # Counter for the statistics
                        packet_counter += 1 # Global counter of packets received by the redvypr instance
                        data_all.append([data,packet_counter])
                    except Exception as e:
                        break
                # Process read packets
                for data_list in data_all:
                    data_packets_fan_out = []
                    data = data_list[0]
                    numpacket = data_list[1]
                    # Add additional information, if not present yet
                    redvypr_packet_statistic.treat_datadict(data, device.name, hostinfo, numpacket, tread,devicedict['devicemodulename'])
                    # Get the devicename
                    raddr = redvypr_address.RedvyprAddress(data)
                    devicename_stat = raddr.address_str
                    # Do statistics
                    try:
                        devicedict['statistics'], status_statistics = redvypr_packet_statistic.do_data_statistics(data, devicedict['statistics'], address_data=raddr)
                        #print('Statistic status',status_statistics)
                    except Exception as e:
                        logger.debug(funcname + ':Statistics:',exc_info=True)

                    #
                    # Check for a command packet
                    #
                    numtag = data['_redvypr']['tag'][hostinfo['uuid']]
                    if numtag < 2:  # Check if data packet fits with addr and its not recirculated again
                        [command, comdata] = data_packets.check_for_command(data, add_data=True)
                        if (command == 'reply'):  # status update
                            device.distribute_data_replyqueue.put_nowait(data)
                        elif (command == 'device_status'):  # status update
                            try:
                                devaddr = comdata['data']['deviceaddr']
                                devstatus = comdata['data']['devicestatus']
                            except:
                                devaddr = None
                                devstatus = None

                            devices_changed.append(device.name) # LEGACY ...
                            if(devaddr is not None):
                                try: # Update the device
                                    devicedict['statistics']['device_redvypr'][devaddr]['_redvypr'].update(devstatus)
                                except Exception as e:
                                    logger.warning('Could not update status ' + str(e))
                                    logger.exception(e)

                            # Send an information about the change, that will trigger a pyqt signal in the main thread
                            devinfo_send = {'type': 'deviceinfo_all', 'deviceinfo_all': copy.deepcopy(deviceinfo_all),
                                            'devices_changed': list(set(devices_changed)), 'device_changed':device.name,
                                            'devices_removed': devices_removed, 'change': 'device_status command','comdata':comdata}
                            infoqueue.put_nowait(devinfo_send)

                    #
                    # Collect the individual dictionaries into one global deviceinfo
                    #
                    try:
                        deviceinfo_all['device_redvypr'][device.name].update(devicedict['statistics']['device_redvypr'])
                    except:
                        deviceinfo_all['device_redvypr'][device.name] = devicedict['statistics']['device_redvypr']

                    # Update metadata
                    if status_statistics['metadata_changed']:
                        try:
                            deviceinfo_all['metadata'][device.name].update(
                                devicedict['statistics']['metadata'])
                        except:
                            deviceinfo_all['metadata'][device.name] = devicedict['statistics']['metadata']

                        # Send a deviceinfo update with the changed metadata
                        compacket = data_packets.commandpacket('info',host=hostinfo,devicename='',packetid='metadata',publisher='')
                        compacket['deviceinfo_all'] = copy.deepcopy(deviceinfo_all)
                        redvypr_packet_statistic.treat_datadict(compacket, '', hostinfo, 0, tread,
                                                                'distribute_data')
                        data_packets_fan_out.append(compacket)

                    #print('Data ready to send',data)
                    #
                    # And finally: Distribute the data
                    #
                    data_packets_fan_out.append(data)
                    # Loop over all devices and check if any subscription works
                    for devicedict_sub in devices:
                        devicesub = devicedict_sub['device']
                        if(devicesub == device): # Not to itself
                            continue

                        for data_packet in data_packets_fan_out:
                            for addr in devicesub.subscribed_addresses: # Loop over all subscribed redvypr_addresses
                                # This is the main functionality for distribution, comparing a datapacket with a
                                # redvypr_address using "in"
                                numtag_packet = data['_redvypr']['tag'][hostinfo['uuid']]
                                #print('Testing packet',redvypr_address.RedvyprAddress(data_packet),numtag_packet,(data_packet in addr))
                                if (data_packet in addr) and (numtag_packet < 2): # Check if data packet fits with addr and if its not recirculated again
                                    try:
                                        #print(funcname + 'data to be sent',data)
                                        devicesub.datainqueue.put_nowait(data_packet) # These are the datainqueues of the subscribing devices
                                        devicedict_sub['statistics']['packets_received'] += 1
                                        #print(devicedict_sub['statistics']['packets_received'])
                                        try:
                                            devicedict_sub['statistics']['packets'][devicename_stat]
                                        except:
                                            devicedict_sub['statistics']['packets'][devicename_stat] = {'received':0,'published':0}
                                        devicedict_sub['statistics']['packets'][devicename_stat]['received'] += 1
                                        #print('Sent data to',devicename_stat,devicedict_sub['packets_received'])
                                        break
                                    except Exception as e:
                                        #logger.exception(e)
                                        thread_status = devicesub.get_thread_status()
                                        if thread_status['thread_running']:
                                            devicedict['statistics']['packets_dropped'] += 1
                                        logger.debug(funcname + ':dataout of :' + devicedict_sub['device'].name + ' full: ' + str(e), exc_info=True)

                    # Fan out the datapacket into the guiqueues of the device
                    for (guiqueue,widget) in devicedict['guiqueues']:  # Put data into the guiqueue, this queue does always exist
                        try:
                            guiqueue.put_nowait(data)
                        except Exception as e:
                            pass
                            # logger.debug(funcname + ':guiqueue of :' + devicedict['device'].name + ' full')

            # Calculate the sleeping time
            tstop = time.time()
            dt_dist = tstop - tstart  # The time for all the looping
            dt_avg += dt_dist
            navg += 1
            # Time to sleep, remove processing time
            dt_sleep = max([0, dt - dt_dist])
            if ((tstop - tinfo) > dt_info):
                tinfo = tstop
                info_dict = {'type':'dt_avg','dt_avg': dt_avg / navg,'packets_processed': packets_processed,'packets_counter':packet_counter,'thread_start':thread_start}
                packets_processed = 0
                # print(info_dict)
                try:
                    infoqueue.put_nowait(info_dict)
                except:
                    pass
        except:
            logger.warning(funcname + 'Could not distribute packets:', exc_info=True)


class Redvypr(QtCore.QObject):
    """This is the redvypr heart. Here devices are added/threads
    are started and data is interchanged.

    """
    device_path_changed = QtCore.pyqtSignal()  # Signal notifying if the device path was changed
    device_added = QtCore.pyqtSignal(list)  # Signal notifying that a device was added
    device_removed = QtCore.pyqtSignal()  # Signal notifying that a device was removed
    devices_connected = QtCore.pyqtSignal(str, str)  # Signal notifying if two devices were connected
    devices_disconnected = QtCore.pyqtSignal(str, str)  # Signal notifying if two devices were connected
    status_update_signal = QtCore.pyqtSignal()  # Signal notifying if the status of redvypr has been changed
    device_status_changed_signal = QtCore.pyqtSignal()  # Signal notifying if datastreams have been added
    hostconfig_changed_signal = QtCore.pyqtSignal()  # Signal notifying if the configuration of the host changed (hostname, hostinfo_opt)

    def __init__(self,config=None,hostname=None,nogui=False,loglevel=None,redvypr_device_scan=None):
        """

        Parameters
        ----------
        config:
            Configuration
        hostname: str
            The hostname of the redvypr instance
        nogui: bool
            No gui if True
        loglevel: logging.loglevel
            The loglevel
        redvypr_device_scan: RedvyprDeviceScan
            External RedvyprDeviceScan object to allow fine grained devices
        """
        super(Redvypr, self).__init__()
        if loglevel is not None:
            logger.setLevel(loglevel)
            logger.debug('Setting loglevel to global: "{}"'.format(loglevel))
        self.__platform__ = __platform__
        funcname = __name__ + '.__init__()'
        logger.debug(funcname)

        if config is None:
            logger.debug(funcname + 'Creating config')
            config = RedvyprConfig(hostname=hostname)
            logger.debug(funcname + 'Config {}'.format(config))

        # print(__platform__)
        # Overwrite hostname with argument
        if hostname is not None:
            config.hostname = hostname

        # global loglevel
        #print('Loglevel',loglevel)
        if loglevel is not None:
            logger.setLevel(loglevel)
            logger.debug('Setting loglevel to global: "{}"'.format(loglevel))
        else: # config loglevel
            try:
                logger.setLevel(config.loglevel)
            except:
                logger.debug('Could not set loglevel to: "{}"'.format(config.loglevel))

        self.hostinfo = create_hostinfo(hostname=config.hostname)
        self.metadata = config.metadata

        self.config = config # This is the initial version
        self.properties = {}  # Properties that are distributed with the device
        self.numdevice = 0
        self.devices = []  # List containing dictionaries with information about all attached devices
        self.device_paths = []  # A list of pathes to be searched for devices
        self.datastreams_dict = {} # Information about all datastreams, this is updated by distribute data
        self.deviceinfo_all = {'device_redvypr':{},'metadata':{}} # Information about all devices, this is updated by distribute data

        self.packets_counter = 0 # Counter for the total number of packets processed
        self.dt_datadist = 0.01  # The time interval of datadistribution
        self.dt_avg_datadist = 0.00  # The time interval of datadistribution
        self.datadistinfoqueue = queue.Queue(maxsize=1000)  # A queue to get informations from the datadistthread
        self.redvyprqueue = queue.Queue()  # A queue to send informations to the datadistthread
        # Lets start the distribution!
        self.datadistthread = threading.Thread(target=distribute_data, args=(
        self.devices, self.hostinfo, self.deviceinfo_all, self.datadistinfoqueue, self.redvyprqueue, self.dt_datadist), daemon=True)
        self.t_thread_start = time.time()
        self.datadistthread.start()

        if redvypr_device_scan is None:
            logger.debug(funcname + ':Searching for devices')
            loglevel_device_scan = logger.getEffectiveLevel()
            #print('Loglevel device scan',loglevel_device_scan,logging.getLevelName(loglevel_device_scan))
            self.redvypr_device_scan = RedvyprDeviceScan(device_path = self.device_paths, redvypr_devices=redvyprdevices, loglevel=loglevel_device_scan)
        else:
            self.redvypr_device_scan = redvypr_device_scan

        logger.debug(funcname + ':Done searching for devices')
        # And now add the devices
        self.add_devices_from_config(config=self.config)
        # A timer to check the status of all threads
        self.devicethreadtimer = QtCore.QTimer()
        self.devicethreadtimer.timeout.connect(self.update_status)  # Add to the timer another update
        self.devicethreadtimer.start(250)

        # A timer to print the status in the nogui environment
        if (nogui):
            self.statustimer = QtCore.QTimer()
            self.statustimer.timeout.connect(self.print_status)
            self.statustimer.start(5000)

    def get_config(self):
        """
        Creates a configuration dictionary out of the current state.

        Returns:
            config: configuration dictionary

        """
        funcname = __name__ + '.get_config():'
        # Devices
        devices = []
        for devicedict in self.devices:
            device = devicedict['device']
            device_config_tmp = device.get_config()
            devices.append(device_config_tmp)

        loglevel_tmp = logging.getLevelName(logger.getEffectiveLevel())
        config = RedvyprConfig(hostname=self.hostinfo['hostname'], metadata=self.metadata,
                               devicepaths=self.device_paths, loglevel=loglevel_tmp,
                               redvyp_version=version, date_created=str(datetime.datetime.utcnow()),
                               devices=devices)

        return config


    def get_devicemodulename_from_str(self, devicename):
        """
        Tries to find a devicemodulename from devicename
        :param devicename:
        :return:
        """
        devicemodulename = None
        funcname = __name__ + '.get_devicemodulename_from_str():'
        logger.debug(funcname)
        #print('Devicename',devicename)
        # Make an exact test first
        for smod in self.redvypr_device_scan.redvypr_devices_flat:
            #print('device:',smod['name'],devicename)
            # The smod['name'] looks like 'redvypr.devices.network.zeromq_device'
            # Check first if devicemodulename has a '.', if not, use split smod['name'] and use the last one
            FLAG_DEVICEMODULENAME_EXACT = False
            if '.' in devicename:
                if (devicename == smod['name']):
                    logger.debug(funcname + ' Found exact fit {}'.format(devicename))
                    FLAG_DEVICEMODULENAME_EXACT = True
                    devicemodulename = smod['name']
                    break
            else:
                smodname = smod['name'].split('.')[-1]
                # print('smodname',smodname)
                if (devicename == smodname):
                    logger.debug(funcname + ' Found exact fit of last entry {}'.format(devicename))
                    FLAG_DEVICEMODULENAME_EXACT = True
                    # device.devicemodulename_orig = device.devicemodulename
                    devicemodulename = smod['name']
                    break

        # Make a test if the string is within the devicemodulename
        if FLAG_DEVICEMODULENAME_EXACT == False:
            logger.debug(funcname + ' Could not add exact fit, searching for substrings')
            for smod in self.redvypr_device_scan.redvypr_devices_flat:
                if (devicename in smod['name']):  # This is a weaker test, can be potentially replaced by regex
                    # device.devicemodulename_orig = device.devicemodulename
                    devicemodulename = smod['name']
                    break

        return devicemodulename

    def add_devices_from_config(self, config):
        funcname = "add_devices_from_config():"
        logger.debug(funcname)
        # Apply the configuration
        if config is not None:
            #print('Config parameter', config, type(config))
            #print('Devicepaths', type(config), type(config.hostname), config.devicepaths)
            # Add device path if found
            devpath = config.devicepaths
            if (type(devpath) == str):
                devpath = [devpath]

            FLAG_NEW_DEVPATH = False
            for p in devpath:
                if (p not in self.device_paths):
                    self.device_paths.append(p)
                    FLAG_NEW_DEVPATH = True

            if FLAG_NEW_DEVPATH:
                self.redvypr_device_scan.scan_devicepath()
                self.device_path_changed.emit()  # Notify about the changes

            # Adding the devices found in the config ['devices']
            # Check if we have a list or something
            #print('devices',config.devices,iter(config.devices))
            try:
                iter(config.devices)
                hasdevices = True
            except:
                hasdevices = False

            # Add the devices
            if (hasdevices):
                for device in config.devices:
                    logger.debug(funcname + 'Adding device {}'.format(device))
                    #print('Devicemodulename:',device.devicemodulename)
                    devicename = device.devicemodulename
                    devicemodulename = self.get_devicemodulename_from_str(devicename)
                    if devicemodulename is not None:
                        device.devicemodulename = devicemodulename
                        logger.info(funcname + 'Adding device {}'.format(device.devicemodulename))
                        #print('-------')
                        #print('Device',device)
                        #print('-------')
                        subscriptions = device.subscriptions
                        dev_added = self.add_device(devicemodulename=device.devicemodulename,
                                                    custom_config=device.custom_config,
                                                    base_config=device.base_config, subscriptions=subscriptions)

                    else:
                        logger.warning(funcname + ' Could not find devicemodulename {}'.format(devicemodulename))

        # Emit a signal that the configuration has been changed
        self.hostconfig_changed_signal.emit()

    def get_deviceinfo(self,publishes=None,subscribes=None):
        """

        Args:
            publishes:
            subscribes:

        Returns:
            A dictionary containing the deviceinfo of all devices seen by this redvypr instance
        """
        if (publishes == None) and (subscribes == None):
            return copy.deepcopy(self.deviceinfo_all)
        else:
            dinfo = {}
            for d in self.devices:
                dev = d['device']
                FLAG_publishes   =   (publishes == dev.publishes) or (publishes == None)
                FLAG_subscribes  = (subscribes == dev.subscribes) or (subscribes == None)
                if FLAG_publishes and FLAG_subscribes:
                    dinfo[dev.name] = copy.deepcopy(dev.statistics['device_redvypr'])

            return dinfo

    def update_status(self):
        funcname = __name__ + '.update_status():'
        # Check if the distribution thread is running, if not warn the user
        if self.datadistthread.is_alive() == False:
            logger.warning('Datadistribution thread is not running! This is bad, consider restarting redvypr.')
            self.status_update_signal.emit()
            if True:
                logger.info('RESTARTING datadistribution thread ....')
                self.datadistthread.start()

            return

        while True:
            try: # Reading data coming from distribute_data thread
                data = self.datadistinfoqueue.get(block=False)
                #print('Got data',data)
                try:
                    if('dt_avg' in data['type']):
                        self.dt_avg_datadist   = data['dt_avg']
                        self.packets_processed = data['packets_processed']
                        self.packets_counter = data['packets_counter']
                        self.t_thread_start = data['thread_start']
                        self.status_update_signal.emit()
                    elif ('deviceinfo_all' in data['type']):
                        data.pop('type')  # Remove the type key
                        # Store the data of the changed devices
                        self.__device_status_changed_data__ = data

                        #print('datastreams changed', data)
                        self.device_status_changed_signal.emit()
                except:
                    logger.info(funcname + 'Error',exc_info=True)

            except Exception as e:
                #logger.exception(e)
                break

    def print_status(self):
        funcname = __name__ + '.print_status():'
        logger.debug(funcname + self.status())

    def status(self):
        """ Creates a statusstr of the devices
        """
        tstr = str(datetime.datetime.now())
        statusstr = "{:s}, {:s}, num devices {:d}".format(tstr, self.hostinfo['hostname'], len(self.devices))

        for sendict in self.devices:
            status = sendict['device'].get_thread_status()
            #info_dict['uuid'] = self.uuid
            #info_dict['thread_uuid'] = self.thread_uuid
            #info_dict['thread_running'] = running
            if(status['thread_running']):
                runstr = 'running'
            else:
                runstr = 'stopped'

            statusstr += '\n\t' + sendict['device'].name + ':' + runstr + ': data packets sent: {:d}' + ': data packets received: {:d}'.format(
                sendict['packets_published'],sendict['packets_received'])
            # statusstr += ': data packets received: {:d}'.format(sendict['numpacketout'])

        return statusstr

    def device_loglevel(self, device, loglevel=None):
        """ Returns the loglevel of the device
        """
        loglevel = 'NA'
        for d in self.devices:
            if d['devices'] == device:
                if (d['logger'] is not None):
                    loglevel = d['logger'].getEffectiveLevel()
                    break

        return loglevel

    def add_device(self, devicemodulename=None, custom_config=None, base_config=None, subscriptions=[]):
        """
        Function adds a device to redvypr

        Args:
            :param devicemodulename:
            :param custom_config: A dictionary with the configuration, this is filled by i.e. a yaml file with the configuration or if clicked in the gui just the name of the device
            :param device_parameter:

        Returns:
            devicelist: a list containing all devices of this redvypr instance


        """
        funcname = self.__class__.__name__ + '.add_device():'
        logger.debug(funcname + ':devicemodule: ' + str(devicemodulename) + ':deviceconfig: ' + str(custom_config))
        #print('add_device custom_config:',custom_config)
        #print('add_device base_config:', base_config)
        devicelist = []
        device_found = False
        # Loop over all modules and check of we find the name
        for smod in self.redvypr_device_scan.redvypr_devices_flat:
            if (devicemodulename == smod['name']):
                logger.debug('Trying to import device {:s}'.format(smod['name']))
                devicemodule = smod['module']
                # Try to get a pydantic base configuration, every device has
                pydantic_base_config = None
                try:
                    pydantic_base_config = devicemodule.DeviceBaseConfig()
                    logger.debug(funcname + ':Found pydantic base configuation {:s}'.format(str(devicemodule)))
                    FLAG_HAS_PYDANTICBASE = True
                    FLAG_PYDANTIC = True
                    # Create or use a given device parameter object
                    #print('type base config',type(base_config))
                    if isinstance(base_config, RedvyprDeviceParameter):
                        #print('Got a device parameter config')
                        device_parameter = base_config
                    elif isinstance(base_config, RedvyprDeviceBaseConfig):
                        #print('Got a base config',base_config)
                        device_parameter = RedvyprDeviceParameter(**base_config.model_dump())
                        #print('parameter',device_parameter)
                    elif isinstance(base_config, dict):
                        #print('Will update from config dictionary')
                        device_parameter_tmp = RedvyprDeviceParameter()
                        device_parameter = device_parameter_tmp.model_copy(update=base_config)
                    else:
                        #print('Standard base_config')
                        device_parameter = RedvyprDeviceParameter()

                    device_parameter.devicemodulename = devicemodulename
                    device_parameter.numdevice = self.numdevice
                    #print('Device parameter',device_parameter)
                    # Update the device parameter with the parameters of the device
                    device_parameter = device_parameter.model_copy(update=pydantic_base_config.model_dump())
                    #print('Device parameter 2', device_parameter)

                except Exception as e:
                    logger.debug(
                        funcname + ':No pydantic base configuration template of device {:s}: {:s}'.format(str(devicemodule), str(e)))
                    # Standard base config with parameter
                    if base_config is not None:
                        device_parameter = RedvyprDeviceParameter(**base_config.model_dump())
                        device_parameter.devicemodulename=devicemodulename
                        device_parameter.numdevice=self.numdevice
                        #print('Device parameter ...',device_parameter)
                    #logger.exception(e)
                    FLAG_HAS_PYDANTICBASE = False
                    FLAG_PYDANTIC = False

                # Try to get a pydantic device specific configuration (the configuration only for the device)
                try:
                    pydantic_device_config = devicemodule.DeviceCustomConfig()
                    #print('Device config of module',pydantic_device_config,type(pydantic_device_config))
                    #print('deviceconfig', custom_config, type(custom_config))

                    try:
                        #print('Custom_Config',custom_config)
                        if custom_config is not None:
                            #print('Dump',custom_config.model_dump())
                            pydantic_custom_config = devicemodule.DeviceCustomConfig.model_validate(custom_config.model_dump())
                        else:
                            pydantic_custom_config = devicemodule.DeviceCustomConfig()
                    except:
                        logger.debug('No config found',exc_info=True)
                        pydantic_custom_config = devicemodule.DeviceCustomConfig()

                    logger.debug(funcname + ':Found pydantic configuration {:s}'.format(str(devicemodule)))
                    #redvypr_device_parameter.config = pydantic_device_config
                    FLAG_HAS_PYDANTIC = True
                except Exception as e:
                    logger.debug(
                        funcname + ':No pydantic configuration template of device {:s}: {:s}'.format(str(devicemodule),
                                                                                            str(e)))
                    #logger.exception(e)
                    FLAG_HAS_PYDANTIC = False
                    pydantic_custom_config = None

                if(device_parameter.maxdevices > 0):
                    ndevices = 0
                    for d in self.devices:
                        devname = d['device'].devicemodulename
                        if(devname == devicemodulename):
                            ndevices += 1

                    if ndevices >= device_parameter.maxdevices:
                        logger.warning(funcname + ' Could not add {:s}, maximum number of {:d} devices reached'.format(devicemodulename,device_parameter.maxdevices))
                        return

                # If the device does not have a name, add a standard but unique one
                devicenames = self.get_all_devicenames()
                #print('Devicenames',devicenames)
                #print('device_parameter',device_parameter)
                #print('devicemodulename', devicemodulename)
                #print('devicename 0',device_parameter.name,len(device_parameter.name))
                devicename_tmp = device_parameter.name
                #print('Devicename_tmp', devicename_tmp)
                if len(device_parameter.name) == 0:
                    logger.debug(funcname + 'using standard name')
                    devicename_tmp = devicemodulename.split('.')[-1]# + '_' + str(self.numdevice)
                    #print('Devicename_tmp',devicename_tmp)

                if devicename_tmp in devicenames:
                    logger.warning(funcname + ' Devicename {:s} exists already, will add {:d} to the name.'.format(devicename_tmp,self.numdevice))
                    devicename_tmp += '_' + str(self.numdevice)

                device_parameter.name = devicename_tmp

                # Check for multiprocess options in configuration
                if 'thread' in device_parameter.multiprocess:  # Thread or QThread
                    dataqueue = queue.Queue(maxsize=queuesize)
                    datainqueue = queue.Queue(maxsize=queuesize)
                    comqueue = queue.Queue(maxsize=queuesize)
                    statusqueue = queue.Queue(maxsize=queuesize)
                    guiqueue = queue.Queue(maxsize=queuesize)
                else: # multiprocess
                    dataqueue = multiprocessing.Queue(maxsize=queuesize)
                    datainqueue = multiprocessing.Queue(maxsize=queuesize)
                    comqueue = multiprocessing.Queue(maxsize=queuesize)
                    statusqueue = multiprocessing.Queue(maxsize=queuesize)
                    guiqueue = multiprocessing.Queue(maxsize=queuesize)

                # guiqueues list
                guiqueues = [[guiqueue, None]]
                # Create a dictionary for the statistics and general information about the device
                # This is used extensively in exchanging information about devices between redvypr instances and or other devices
                statistics = redvypr_packet_statistic.create_data_statistic_dict()
                # Device do not necessarily have a statusqueue
                try:
                    numdevices = len(self.devices)
                    # Could also create a UUID based on the hostinfo UUID str
                    device_parameter.uuid = '{:03d}--'.format(numdevices) + str(uuid.uuid1()) + '::' + self.hostinfo['uuid']
                    try:
                        devicemodule.Device
                        HASDEVICE = True
                    except:
                        HASDEVICE = False

                    if (HASDEVICE):  # Module has its own device
                        Device = devicemodule.Device
                        # Check if a startfunction is implemented
                        # TODO: This seems to be bogus, double check
                        try:
                            Device.start
                            startfunction = None
                            logger.debug(
                                funcname + ' Device has a start function')
                        except:
                            logger.debug(
                                funcname + ' Device does not have a start function, will add the standard function')
                            startfunction = devicemodule.start
                    else:
                        Device = RedvyprDevice
                        startfunction = devicemodule.start

                    # Config used at all?
                    #print('Getting config')

                    #print('Done')


                    if FLAG_HAS_PYDANTIC:
                        logger.debug(funcname + ' Using pydantic config')

                    logger.debug(funcname + 'Custom config for device')
                    logger.debug(funcname + 'Config: {:s}'.format(str(pydantic_custom_config)))
                    # Set the loglevel
                    try:
                        level = custom_config['loglevel']
                    except Exception as e:
                        logger.info(funcname + 'Setting loglevel to standard')
                        # Set the loglevel
                        level = logger.getEffectiveLevel()

                    levelname = logging.getLevelName(level)
                    logger.debug(funcname + 'Setting the loglevel to {}'.format(levelname))
                    device_parameter.loglevel = levelname
                    # Creating the device
                    #print('Deviceparameter',device_parameter)
                    device = Device(device_parameter=device_parameter, custom_config=pydantic_custom_config,
                                    redvypr=self, dataqueue=dataqueue,
                                    comqueue=comqueue, datainqueue=datainqueue,
                                    statusqueue=statusqueue, guiqueues=guiqueues,
                                    statistics=statistics, startfunction=startfunction)

                    # Subscribe to info packets from redvypr itself
                    device.subscribe_address(metadata_address)
                    device.subscription_changed_signal.connect(self.process_subscription_changed)
                    self.numdevice += 1
                    # If the device has a logger
                    devicelogger = device.logger
                except Exception as e:
                    logger.warning(funcname + ' Could not add device because of:')
                    logger.exception(e)


                devicedict = {'device':device, 'guiqueues': guiqueues,
                              'statistics': statistics, 'logger': devicelogger,'comqueue':comqueue}

                # Add the modulename
                devicedict['devicemodulename'] = devicemodulename
                # Add some statistics (LEGACY)
                devicedict['packets_received'] = 0
                devicedict['packets_published'] = 0
                # The displaywcreate_idget, to be filled by redvyprWidget.add_device (optional)
                devicedict['devicedisplaywidget'] = None
                # device = devicedict['device']

                # Check if the device wants a direct start after initialization
                try:
                    autostart = device.autostart
                except:
                    autostart = False

                # Add the device to the device list
                self.devices.append(devicedict)  # Add the device to the devicelist
                ind_device = len(self.devices) - 1

                # Update the statistics of the device itself
                deviceinfo_packet = {'_redvypr':{},'_deviceinfo': {'subscribes': device_parameter.subscribes, 'publishes': device_parameter.publishes, 'devicemodulename': device_parameter.devicemodulename}}
                device.dataqueue.put(deviceinfo_packet)
                # Send a device_status packet to notify that a new device was added (deviceinfo_all) is updated
                datapacket = data_packets.commandpacket(command='device_status')
                device.dataqueue.put(datapacket)
                #print('Autostart autostart',autostart)
                if (autostart):
                    logger.info(funcname + ': Starting device')
                    self.start_device_thread(device)

                devicelist = [devicedict, ind_device, devicemodule]

                #
                # Subscribe to devices
                #
                try:
                    subscribe_addresses = subscriptions
                except:
                    logger.debug(funcname + ' No subscriptions found')
                    subscribe_addresses = []

                logger.debug(funcname + 'Subscribing to')
                for a in subscribe_addresses:
                    logger.debug(funcname + 'subscribing: {}'.format(a))
                    device.subscribe_address(a)

                logger.debug(funcname + ': Emitting device signal')
                self.device_added.emit(devicelist)
                device_found = True
                break

        if (device_found == False):
            logger.warning(funcname + ': Could not add device (not found): {:s}'.format(str(devicemodulename)))

        #return devicelist
        return device

    def start_device_thread(self, device):
        """ Functions starts the device thread that is the core of each device
        Args:
           device: Device is either a Device class defined in the device module or a dictionary containing the device class (when called from infodevicewidget in redvypr.py)

        """
        funcname = __name__ + '.start_device_thread():'
        if (type(device) == dict):  # If called from devicewidget
            device = device['device']

        # logger.debug(funcname + 'Starting device: ' + str(device.name))
        logger.debug(funcname + 'Starting device: ' + str(device.name))
        thread = device.thread_start()

    def stop_device_thread(self, device):
        """Functions stops a thread, to process the data (i.e. reading from a device, writing to a file)
        Args:
           device: Device is either a Device class defined in the device module or a dictionary containing the device class (when called from infodevicewidget in redvypr.py)

        """
        funcname = __name__ + '.stop_device_thread()'
        command = data_packets.commandpacket(command='stop', uuid=device.uuid)
        if (type(device) == dict):  # If called from devicewidget
            device = device['device']
        logger.debug(funcname + ':Stopping device: ' + device.name)
        thread_status = device.thread_stop()

        for sendict in self.devices:
            if (sendict['device'] == device):
                if (sendict['device'].thread == None):
                    return
                elif (sendict['device'].thread_running()):
                    try:
                        device.thread_stop()
                        return
                    except:
                        pass
                    try:
                        if (device.datainqueuestop):
                            datainqueuestop = True
                    except:
                        datainqueuestop = False

                    device.thread_stop()

                else:
                    logger.warning(funcname + ': Could not stop thread.')

    def adddevicepath(self, folder):
        """Adds a path to the devicepathlist
        """
        if folder:
            if (folder not in self.device_paths):
                self.device_paths.append(folder)
                #print('scanning')
                #self.redvypr_device_scan.logger.setLevel(logging.DEBUG)
                self.redvypr_device_scan.scan_devicepath()
                self.device_path_changed.emit()  # Notify about the changes

    def remdevicepath(self, folder):
        if (folder not in self.device_paths):
            self.device_paths.remove(folder)
            self.device_path_changed.emit()  # Notify about the changes

    def process_subscription_changed(self):
        """
        Process the subscription changed signals of the devices
        Returns:

        """
        devsender = self.sender()
        logger.debug('Subscribtion changed {}'.format(devsender.name))
        for d in self.devices:
            dev = d['device']
            if dev == devsender:
                continue
            dev.subscription_changed_global(devsender)

    def get_all_subscriptions(self):
        """

        Returns:
            List containing two lists of equal length, the first with all subscriptions, the second the device address

        """
        all_subscriptions = []
        all_devices = []
        for d in self.devices:
            dev = d['device']
            nsub = len(dev.subscribed_addresses)
            all_subscriptions.extend(dev.subscribed_addresses)
            all_devices.extend([dev.address_str]*nsub)

        return [all_subscriptions,all_devices]

    def get_all_devicenames(self):
        """
        Returns a list with the devicenames
        """
        devicenames = []
        for d in self.devices:
            devicenames.append(d['device'].name)

        return devicenames

    def get_hosts(self):
        """
        Returns: List with all redvypr hosts known to this redvypr host instance

        """
        hosts = []
        for d in self.devices:
            dev = d['device']
            hosts.extend(dev.get_hosts())

        hosts = list(set(hosts))
        hosts.sort()
        return hosts


    def get_device_objects(self, publishes=None, subscribes=True):
        """
        Returns a list of all devices of this redvypr instance. Returns all devices if neither "publishes" or "subscribes" is defined.
        Args:
            publishes: True/False or None
            subscribes: True/False

        Returns: List with redvypr_devices

        """

        devicelist = []
        for d in self.devices:
            dev = d['device']
            if publishes == True:
                if(dev.publishes):
                    devicelist.append(dev)
                    continue
            elif publishes is None:
                devicelist.append(dev)
                continue

            if (subscribes):
                if (dev.subscribes):
                    devicelist.append(dev)
                    continue

        return devicelist

    def get_devices(self, local=None, local_object=True):
        """
        Returns a list of all devices of this redvypr instance.
        Returns: List with devicenames

        """

        devicelist = []
        for d in self.devices:
            dev = d['device']
            if local_object:
                devicelist.append(dev.name)
            else:
                devaddrs = dev.get_deviceaddresses(local=local)
                for devaddr in devaddrs:
                    devicelist.append(devaddr.devicename)

        devicelist = list(set(devicelist))
        devicelist.sort()
        return devicelist

    def get_deviceaddresses(self, local = None, publishes = None, subscribes = None):
        """
        Returns a list of redvypr_addresses of all devices. If local == None all known devices are listed,
        also of all remote devices that are forwarded by a local host device (i.e. iored device).


        Args:
            local [bool/None]: None: all known devices are listed, False: Remote devices are listed, True: local devices are listed

        Returns:
            List of redvypr_address


        """
        raddrs = []
        for dev in self.devices:
            raddrs_tmp = dev['device'].get_deviceaddresses(local)
            raddrs.extend(raddrs_tmp)

        return raddrs

    def get_datakeys(self, local=None):
        """
        Returns a list of all datakeys this host is providing by all its devices
        Returns:
            List of datakeys (str)
        """
        datakeys = []
        for dev in self.devices:
            dkeys = dev['device'].get_datakeys()
            datakeys.extend(dkeys)

        datakeys = list(set(datakeys))
        datakeys.sort()
        return datakeys

    def get_datastreams(self,local=None):
        """

        Args:
            local:

        Returns:

        """
        datastreams = []
        for dev in self.devices:
            raddrs_tmp = dev['device'].get_datastreams()
            datastreams.extend(raddrs_tmp)

        return datastreams

    def get_packetids(self):
        """
        Returns a list of all packetids this host has been seen
        Returns:
            List of packetids (str)
        """
        packetids = []
        for dev in self.devices:
            dkeys = dev['device'].get_packetids()
            packetids.extend(dkeys)

        packetids = list(set(packetids))
        packetids.sort()
        return packetids

    def get_metadata(self, address, publisher_strict=False, mode='merge'):
        funcname = __name__ + 'get_metadata():'
        logger.debug(funcname)
        deviceinfo_all = self.get_deviceinfo()
        metadata = redvypr_packet_statistic.get_metadata_deviceinfo_all(deviceinfo_all, address=address, publisher_strict=publisher_strict, mode=mode)
        return metadata

    def get_metadata_commandpacket(self):
        funcname = __name__ + 'get_metadata_commandpacket():'
        logger.debug(funcname)
        deviceinfo_all = self.get_deviceinfo()
        compacket = data_packets.commandpacket('info', host=self.hostinfo, devicename='', packetid='metadata', publisher='')
        compacket['deviceinfo_all'] = copy.deepcopy(deviceinfo_all)
        tread = time.time()
        redvypr_packet_statistic.treat_datadict(compacket, '', self.hostinfo, 0, tread,
                                                'distribute_data')
        return compacket

    def get_known_devices(self):
        """ List all known devices that can be loaded by redvypr
        
        Returns:
        --------
        list
            A list of known devices
        """
        funcname = self.__class__.__name__ + '.get_known_devices():'
        logger.debug(funcname)
        devices = []
        for d in self.redvypr_device_scan.redvypr_devices_flat:
            devices.append(d['name'])

        return devices

    def rem_device(self,device):
        """
        Removes a device, after the removal a signal for device change is sent
        Args:
            device: redvypr_device

        Returns:

        """
        funcname = self.__class__.__name__ + '.rem_device():'
        logger.debug(funcname +'Removing device:{}'.format(device.name))
        FLAG_REMOVED = False
        # Search for the device in sendict
        for sendict in self.devices:
            if(sendict['device'] == device):
                #print('Sendict',sendict)
                if (sendict['device'].thread == None):
                    logger.debug(funcname + 'Thread is not running, doing nothing')
                    pass
                elif (sendict['device'].thread_running()):
                    logger.debug(funcname + 'Sending stop command')
                    device.thread_stop()

                self.devices.remove(sendict)
                FLAG_REMOVED = True
                self.device_removed.emit()
                device_changed_dict = {'type':'device_removed','device':device.name,'uuid':device.uuid}
                self.redvyprqueue.put(device_changed_dict)
                break

        return FLAG_REMOVED

















"""
network device
"""


import datetime
import logging
import queue
from PyQt5 import QtWidgets, QtCore, QtGui
import time
import numpy as np
import logging
import sys
import threading
import socket
#from apt_pkg import config
import yaml
import copy
import pydantic
import typing
from redvypr.device import RedvyprDevice
from redvypr.data_packets import check_for_command
import redvypr.redvypr_address as redvypr_address
import redvypr.packet_statistic as packet_statistic
from redvypr.widgets.pydanticConfigWidget import dictQTreeWidget
from redvypr.widgets.standard_device_widgets import displayDeviceWidget_standard

logging.basicConfig(stream=sys.stderr)
logger = logging.getLogger('network_device')
logger.setLevel(logging.DEBUG)

redvypr_devicemodule = True

class DeviceBaseConfig(pydantic.BaseModel):
    publishes: bool = True
    subscribes: bool = True
    description: str = 'Send and receive data using standard network protocols as TCP or UDP'
    gui_tablabel_display: str = 'Network Status'

class DeviceCustomConfig(pydantic.BaseModel):
    address: str = pydantic.Field(default='<IP>', description='The IP address, this can be also <IP> or <broadcast>')
    port: int = pydantic.Field(default=18196, description='The network port used')
    protocol: typing.Literal['tcp', 'udp'] = pydantic.Field(default='tcp', description= 'The network protocol used.')
    direction: typing.Literal['publish', 'receive'] = pydantic.Field(default='tcp', description='Publishing or receiving data.')
    datakey: str = pydantic.Field(default='all', description='Datakey to store data, this is used if serialize is raw or str')
    serialize: typing.Literal['yaml','str','raw'] = pydantic.Field(default='raw',description='Method to serialize (convert) original data into binary data.')
    queuesize: int = pydantic.Field(default=10000, description= 'Size of the queues for transfer between threads')
    dt_status: float = pydantic.Field(default=4.0,description= 'Send a status message every dt_status seconds')
    tcp_reconnect: bool = pydantic.Field(default=True, description = 'Reconnecting to TCP Port if connection was closed by host')
    tcp_numreconnect: int = pydantic.Field(default= 10, description='The number of reconnection attempts before giving up')


#https://stackoverflow.com/questions/166506/finding-local-ip-addresses-using-pythons-stdlib
def get_ip():
    s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    try:
        # doesn't even have to be reachable
        s.connect(('10.255.255.255', 1))
        IP = s.getsockname()[0]
    except Exception:
        IP = '127.0.0.1'
    finally:
        s.close()
    return IP


def yaml_dump(data):
    #return yaml.dump(data,default_flow_style=False)
    return yaml.dump(data,explicit_end=True,explicit_start=True)

def raw_to_packet(datab, config, safe_load=True):
    """
    Packs the received raw data into a packet that can be sent via the dataqueue

    Args:
        datab:
        config:
        safe_load:

    Returns:
        packets: A list of datapackets to be sent via the dataqueue

    """
    funcname = __name__ + '.raw_to_packet()'
    t = time.time()
    packets = []
    datab_rest = b''
    #print('A data:',datab)
    #print('-------------------')
    if safe_load:
        loader = yaml.SafeLoader
    else:
        loader = yaml.CLoader
    if(config['serialize'] == 'yaml'):
        i0 = datab.find(b'---\n')
        i1 = datab.rfind(b'...\n')+4
        # Check here if we got some data
        datab_valid = datab[i0:i1]
        datab_rest = datab[i1:]
        for databs in datab_valid.split(b'...\n'): # Split the text into single subpackets
            try:
                data = yaml.load(databs,Loader=loader)
            except Exception as e:
                #print('data',databs)
                logger.info(funcname + ': Could not decode message:',exc_info=True)
                logger.debug(funcname + ': Could not decode message  with supposed format {:s} into something useful.'.format(config['datakey']))
                data = None
            if(data is not None):
                if(config['datakey'] == 'all'): # Forward the whole message
                    packets.append(data)
                else:
                    datan = {'t':t}
                    datan[config['datakey']] = data[config['datakey']]
                    packets.append(datan)

    elif (config['serialize'] == 'utf-8'):  # Put the "str" data into the packet with the key in "data"
        data = datab.decode('utf-8')
        datan = {'t':t}
        datan[config['datakey']] = data
        packets.append(datan)
    elif(config['serialize'] == 'raw'): # Put the "str" data into the packet with the key in "data"
        datan = {'t':t}
        datan[config['datakey']] = datab
        packets.append(datan)

    return {'packets':packets,'datab_rest':datab_rest}

def packet_to_raw(data_dict,config):
    """ Function that processes the data_dict to a serialized datastream sendable over network connections 
    """
    funcname = __name__ + 'send_data()'
    #
    if (config['datakey'] == 'all') and (config['serialize'] == 'yaml'):
        datab = yaml_dump(data_dict).encode('utf-8')
    elif (config['serialize'] == 'utf-8'):
        key = config['datakey']
        datab = str(data_dict[key]).encode('utf-8')
    elif (config['serialize'] == 'raw'):
        key = config['datakey']
        data = data_dict[key]
        if(isinstance(data, (bytes, bytearray))):
            datab = data
        else:
            datab = str(data).encode('utf-8')

    return datab


def handle_tcp_client_send(client, threadqueue, statusqueue, config):
    funcname =  __name__ + 'handle_tcp_client_send()'
    address = client[1]
    client = client[0]
    statusstr = 'Sending data to (TCP):' + str(address)
    nsent = 0
    dt_status = config['dt_status']
    tlast_status = time.time()
    #statusdict = {'tcp': {'send': {'address': {address: {'sent': nsent, 'status': 'open'}}}}}
    try:
        statusqueue.put_nowait(statusstr)
        #statusqueue.put_nowait(statusdict)
        logger.info(funcname + ':' + statusstr)
    except Exception as e:
        logger.warning(funcname + ':' + str(e))
    
    while True:
        data = threadqueue.get() # Blocking
        tread = time.time()
        #print('Read data from queue')
        try:
            logger.debug(funcname  + ':Sending data: {:s}'.format(str(data)))
            client.send(data)
            nsent += len(data)
        except Exception as e:
            statusstr = 'Connection to {:s} closed, stopping thread, sent {:d}bytes'.format(str(address),nsent)
            try:
                statusqueue.put_nowait(statusstr)
            except: # If the queue is full
                pass
            logger.info(funcname + ':' + statusstr)
            return

        if (tread - tlast_status) > dt_status:
            statusstr = 'Connection to {:s} open, sent {:d}bytes'.format(str(address), nsent)
            #statusdict = {'tcp': {'send': {'address': {address:{'sent':nsent,'status':'open'}}}}}
            try:
                statusqueue.put_nowait(statusstr)
                #statusqueue.put_nowait(statusdict)
            except:  # If the queue is full
                pass
            logger.debug(funcname + ':' + statusstr)
            tlast_status = time.time()



def start_tcp_send(dataqueue, datainqueue, statusqueue, config=None, device_info=None):
    """ TCP publishing, this thread waits for TCP connections connections. For each connection a new connection is created that is fed with a subthread (handle_tcp_send) is created that is receiving data with a new queue. 
    """
    funcname = __name__ + '.start_tcp_send()'
    logger.debug(funcname + ':Starting network thread')        
    npackets     = 0 # Number of packets
    bytes_read   = 0
    metadata_packet = None
    metadata_update = False
    threadqueues = []
    server = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    logger.debug(funcname + ' Binding to {:s}:{:d}'.format(config['address'],config['port']))
    server.bind((config['address'],config['port']))
    server.listen(8)
    server.settimeout(0.05) # timeout for listening
    # Some variables for status
    # Some variables for status
    try:
        dt_status = config['dt_status']
    except:
        dt_status = 2  # Send a status message every dtstatus seconds

    # Adding dt_status to get first status asap
    tstatus = time.time() - dt_status + 0.1
    npackets = 0 # Number packets received via the datainqueue
    clients = []
    try:
        queuesize = config['queuesize']
    except:
        queuesize = 1000

    FLAG_RUN = True
    while FLAG_RUN:
        try:
            client, address = server.accept() # Here the timeout will let the loop run with 20 Hz
            taccept = time.time()
            threadqueue = queue.Queue(maxsize=queuesize)
            tcp_thread = threading.Thread(target=handle_tcp_client_send, args=([client,address],threadqueue,statusqueue,config), daemon=True)
            tcp_thread.start()
            clients.append(client)
            thread_queue_dict = {'thread': tcp_thread, 'queue': threadqueue, 'address': address, 'bytes_sent': 0, 'packets_published': 0,
             'taccept': taccept}
            threadqueues.append(thread_queue_dict)
            if metadata_packet is not None:
                metadata_send = copy.deepcopy(metadata_packet)
                devinfo_all = metadata_send.pop('deviceinfo_all')
                metadata_tmp = devinfo_all['metadata']
                metadata_send['_metadata'] = {}
                raddr_tmp = redvypr_address.RedvyprAddress(metadata_packet,datakey='')
                raddr_tmp_str = raddr_tmp.get_fullstr()
                raddr_tmp_str = 'REMOTE__' + raddr_tmp_str
                metadata_send['_metadata'][raddr_tmp_str] = metadata_tmp
                datab = packet_to_raw(metadata_send, config)
                q = thread_queue_dict
                q['queue'].put(datab)
                q['bytes_sent'] += len(datab)
                q['packets_published'] += 1


        except socket.timeout:
            pass
        except:
            logger.info(funcname + ':thread start: ',exc_info=True)
        
        while(datainqueue.empty() == False):
            try:
                data_dict = datainqueue.get(block=False)
                [command, comdata] = check_for_command(data_dict, thread_uuid=device_info['thread_uuid'],
                                                                    add_data=True)
                packet_address = redvypr_address.RedvyprAddress(data_dict)
                if (command is not None):
                    logger.debug('Command is for me: {:s}'.format(str(command)))
                    if(command == 'stop'):
                        logger.info(funcname + 'received command:' + str(data_dict) + ' stopping now')
                        logger.debug('Stop command')
                        for client in clients:
                            client.close()
                        server.close()
                        FLAG_RUN = False
                        break

                    if (command == 'info'):
                        logger.debug('Metadata command')
                        if packet_address.packetid == 'metadata':
                            metadata_packet = data_dict
                            metadata_update = True

                npackets += 1
                # Call the send_data function to create a binary sendable datastream
                datab = packet_to_raw(data_dict,config)
                #print('sending data',datab)
                for q in threadqueues:
                    q['queue'].put(datab)
                    q['bytes_sent'] += len(datab)
                    q['packets_published'] += 1

            except Exception as e:
                logger.debug(funcname + ':Exception:' + str(e))



        # Sending a status message
        if((time.time() - tstatus) > dt_status):
            statusdata = {'npackets':npackets}
            tstatus = time.time()
            # Do a cleanup of the threads that are dead
            for q in threadqueues:
                isalive = q['thread'].is_alive() # This can be False or None if the thread is dead already
                if(isalive == False):
                    threadqueues.remove(q)
                
            statusdata['tcp_clients'] = []
            #statusdata['config'] = copy.deepcopy(config)
            for q in threadqueues:
                taccept_str = datetime.datetime.fromtimestamp(q['taccept'])
                client = {'bytes':q['bytes_sent'],'address':q['address'][0],'port':q['address'][1],'packets':q['packets_published'],'t_accept':taccept_str}
                statusdata['tcp_clients'].append(client)
                
            try:
                statusqueue.put_nowait(statusdata)
            except: # If the queue is full
                pass                

            
        
        
            
def start_tcp_recv(dataqueue, datainqueue, statusqueue, config=None, device_info=None):
    """ TCP receiving
    """
    funcname = __name__ + '.start_tcp_recv():'
    logger.debug(funcname + ':Starting network thread')        
    sentences = 0
    bytes_read = 0
    reconnections = 0
    threadqueues = []
    client = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    logger.debug(funcname + ': Connecting to '+ str(config))
    client.settimeout(1.0) # timeout for listening
    try:
        client.connect((config['address'],config['port']))
    except Exception as e:
        logger.warning(funcname + ': Could not connect to host. {:s}'.format(str(e)))
        return
    
    client.settimeout(0.05) # timeout for listening
    
    # Some variables for status
    try:
        dt_status = config['dt_status']
    except:
        dt_status = 2 # Send a status message every dtstatus seconds

    # Adding dt_status to get first status asap
    tstatus = time.time() - dt_status + 0.1
    # Reconnecting to TCP Port if connection was closed by host
    try:
        config['tcp_reconnect']
    except:
        config['tcp_reconnect'] = True
        
    # The number of reconnection attempts before giving up
    try:
        config['tcp_numreconnect']
    except:
        config['tcp_numreconnect'] = 10
        
    # 
    npackets = 0 # Number packets received via the datainqueue
    datab_all = b''
    while True:
        try:
            com = datainqueue.get(block=False)
            [command, comdata] = check_for_command(com, thread_uuid=device_info['thread_uuid'],
                                                   add_data=True)
            if (command is not None):
                logger.debug(funcname + 'Command is for me: {:s}'.format(str(command)))
                if (command == 'stop'):
                    logger.info(funcname + 'received command:' + str(command) + ' stopping now')
                    client.close()
                    #FLAG_RUN = False
                    break
        except:
            pass

        try:
            datab = client.recv(1000000)
            if(datab == b''): # Connection closed
                logger.warning(funcname + ': Connection closed by host')
                if(config['tcp_reconnect']): # Try to reconnect
                    logger.warning(funcname + ': Reconnecting')
                    i = 0
                    while i < config['tcp_numreconnect']: # a couple of times with a 1 second interval
                        i = i + 1
                        try:
                            client = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
                            client.settimeout(1.0) # timeout for listening
                            client.connect((config['address'],config['port']))
                            reconnections += 1
                        except:
                            logger.warning(funcname + ': Could not connect to host.')
                            
                    logger.info(funcname + ': Connected to '+ str(config))
                    client.settimeout(0.05) # timeout for listening
                else:
                    logger.warning(funcname + ': Stopping thread')
                    client.close()
                    break

            bytes_read += len(datab)
            # Check what data we are expecting and convert it accordingly
            datab_all += datab
            tmp = raw_to_packet(datab_all, config, safe_load=False)
            packets = tmp['packets']
            datab_all = tmp['datab_rest']
            for p in packets:
                ## Check if there is a deviceinfo command
                #command = check_for_command(p)
                #p_address=redvypr_address.RedvyprAddress(p)
                #if (command == 'info'):
                #    logger.debug('Metadata command')
                #    if p_address.packetid == 'metadata':

                print(funcname + ' publishing packet',p)
                dataqueue.put(p)
                npackets += 1

        except socket.timeout as e:
            pass
        except:
            logger.info(funcname + ':',exc_info=True)
            return

        # Sending a status message
        if((time.time() - tstatus) > dt_status):
            statusdata = {}
            statusdata['bytes_read'] = bytes_read
            statusdata['time'] = str(datetime.datetime.now())
            statusdata['reconnections'] = reconnections
            tstatus = time.time()
            #statusdata['clients'] = []
            statusdata['config'] = copy.deepcopy(config)
                
            try:
                statusqueue.put_nowait(statusdata)
            except: # If the queue is full
                pass

def start_udp_send(dataqueue, datainqueue, statusqueue, config=None, device_info=None):
    """ UDP publishing
    """
    funcname = __name__ + '.start_udp_send()'
    logger.debug(funcname + ':Starting network thread')        
    sentences = 0
    bytes_sent = 0

    # Some variables for status
    tstatus = time.time()
    dt_status = 5 # Send a status message every dtstatus seconds
    npackets = 0 # Number packets received via the datainqueue
    client = socket.socket(socket.AF_INET, socket.SOCK_DGRAM, socket.IPPROTO_UDP) # UDP
    if(config['address'] == '<broadcast>'):
        #client.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEPORT, 1)
        client.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR,  1)  # https://stackoverflow.com/questions/13637121/so-reuseport-is-not-defined-on-windows-7
        # Enable broadcasting mode
        client.setsockopt(socket.SOL_SOCKET, socket.SO_BROADCAST, 1)
        udp_addr = ""
    else:
        udp_addr = config['address']

    client.settimeout(0.01) # timeout for listening, do we need that here??
    FLAG_RUN=True
    while FLAG_RUN:
        time.sleep(0.05)
        while(datainqueue.empty() == False):
            try:
                data_dict = datainqueue.get(block=False)
                command = check_for_command(data_dict, thread_uuid=device_info['thread_uuid'])
                if (command is not None):
                    logger.debug('Command is for me: {:s}'.format(str(command)))
                    client.close()
                    logger.info(funcname + 'received command:' + str(data_dict) + ' stopping now')
                    statusdata = {}
                    udpaddrstr = str((udp_addr, config['port']))
                    statusdata = '{} Stopping UDP send thread to {}'.format(str(datetime.datetime.now()),udpaddrstr)
                    try:
                        statusqueue.put_nowait(statusdata)
                    except:  # If the queue is full
                        pass

                    FLAG_RUN = False
                    break
                else:
                    npackets   += 1
                    # Call the send_data function to create a binary sendable datastream
                    datab       = packet_to_raw(data_dict,config)
                    bytes_sent += len(datab)
                    #print('Sending data',datab,udp_addr,config['port'])
                    client.sendto(datab, (udp_addr, config['port']))


            except Exception as e:
                logger.debug(funcname + ':Exception:' + str(e))



        # Sending a status message
        if((time.time() - tstatus) > dt_status):
            statusdata = {'npackets':npackets}
            statusdata['bytes_sent'] = bytes_sent
            statusdata['udp_address'] = str((udp_addr, config['port']))
            tstatus = time.time()
            #statusdata['clients'] = []
            #statusdata['config'] = copy.deepcopy(config)
            statussend = {}
            statussend['udp_publish'] = statusdata

            try:
                statusqueue.put_nowait(statussend)
            except: # If the queue is full
                pass
            


        
        
def start_udp_recv(dataqueue, datainqueue, statusqueue, config=None, device_info = None):
    """ UDP receiving
    """
    funcname = __name__ + '.start_udp_recv():'
    logger.debug(funcname + ':Starting network thread (uuid: {:s})'.format(device_info['thread_uuid']))
    npackets     = 0
    bytes_read   = 0
    threadqueues = []
    tstatus      = 0
    dt_status     = 2 # Send a status message every dtstatus seconds
    
    #client = socket.socket(socket.AF_INET, socket.SOCK_DGRAM, socket.IPPROTO_UDP) # UDP
    client = socket.socket(socket.AF_INET, socket.SOCK_DGRAM) # UDP
    if(config['address'] == '<broadcast>'):
        #client.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEPORT, 1) # https://stackoverflow.com/questions/13637121/so-reuseport-is-not-defined-on-windows-7
        client.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)  # https://stackoverflow.com/questions/13637121/so-reuseport-is-not-defined-on-windows-7
        # Enable broadcasting mode
        client.setsockopt(socket.SOL_SOCKET, socket.SO_BROADCAST, 1)
        udp_addr = ""
    else:
        udp_addr = config['address']
    client.settimeout(0.05) # timeout for listening
    logger.debug(funcname + 'Will bind to {:s} on port {:d}'.format(udp_addr,config['port']))
    client.bind((udp_addr,config['port']))
    while True:
        try:
            com = datainqueue.get(block=False)
            command = check_for_command(com,thread_uuid=device_info['thread_uuid'])
            logger.debug('Got a command: {:s}'.format(str(com)))
            if (command is not None):
                logger.debug('Command is for me: {:s}'.format(str(command)))
                client.close()
                logger.info(funcname + 'received command:' + str(com) + ' stopping now')
                statusdata = {}
                statusdata['status'] = 'Stopping UDP redcv thread'
                statusdata['time'] = str(datetime.datetime.now())
                try:
                    statusqueue.put_nowait(statusdata)
                except:  # If the queue is full
                    pass

                break

        except:
            pass

        try:
            datab, addr = client.recvfrom(1000000)
            #print('Got data',datab,addr)
            bytes_read += len(datab)
            t = time.time()
            # Check what data we are expecting and convert it accordingly
            packets = raw_to_packet(datab, config, safe_load=False)
            for p in packets:
                dataqueue.put(p)
                npackets += 1

            #print(config)
            #print('UDP Received',datab)
            #print('UDP sent as',datan)
        except socket.timeout as e:
            pass
        except Exception as e:
            logger.info(funcname + ': Error: ' + str(e))
            logger.info(funcname + ': Will stop now')
            return


        # Sending a status message
        if((time.time() - tstatus) > dt_status):
            statusdata = {}
            statusdata['bytes_read'] = bytes_read
            statusdata['npackets'] = npackets
            statusdata['time'] = str(datetime.datetime.now())            
            tstatus = time.time()
            #statusdata['clients'] = []
            statusdata['config'] = copy.deepcopy(config)
                
            try:
                statusqueue.put_nowait(statusdata)
            except: # If the queue is full
                pass


    logger.info(funcname + ' stopped')



def start(device_info, config, dataqueue, datainqueue, statusqueue):
    """

    Args:
        device_info:
        config:
        dataqueue:
        datainqueue:
        statusqueue:

    Returns:

    """
    funcname = __name__ + '.start():'
    logger.debug(funcname + config['protocol'])
    if (config['address'].lower() == '<ip>'):
        ip = get_ip()
        logger.debug(funcname + ' Converting <IP> to {:s}'.format(ip))
        config['address'] = ip

    if(config['direction'] == 'publish'):
        if(config['protocol'] == 'tcp'):
            logger.info(__name__ + ':Start to serve data on address (TCP):' + str(config))
            start_tcp_send(dataqueue,datainqueue,statusqueue,config=config,device_info=device_info)
        elif(config['protocol'] == 'udp'):
            logger.info(__name__ + ':Start to serve data on address (UDP)' + str(config))
            #start_udp_send(dataqueue,datainqueue,statusqueue,config=config)
            start_udp_send(dataqueue, datainqueue, statusqueue, config=config,
                           device_info=device_info)
    elif(config['direction'] == 'receive'):
        if(config['protocol'] == 'tcp'):
            logger.info('Start to receive data from address (TCP):' + str(config))
            start_tcp_recv(dataqueue,datainqueue,statusqueue,config=config,device_info=device_info)
        elif(config['protocol'] == 'udp'):
            logger.info('Start to receive data from address (UDP):' + str(config))
            start_udp_recv(dataqueue,datainqueue,statusqueue,config=config,device_info=device_info)

class Device(RedvyprDevice):
    network_status_changed = QtCore.pyqtSignal()  # Signal notifying that the network status changed
    def __init__(self, **kwargs):
        """
        """
        super(Device, self).__init__(**kwargs)
        self.check_and_fill_config() # Add standard stuff
        # Could be done with a blocking thread, but lets keep it simple
        self.network_status = {}
        self.statustimer = QtCore.QTimer()
        self.statustimer.timeout.connect(self.get_status)  # Add to the timer another update
        self.statustimer.start(1000)

    def get_metadata(self, address, mode='merge'):
        """
        Gets the metadata of the redvypr address
        :param address:
        :return:
        """
        funcname = __name__ + '.get_metadata({},{}):'.format(str(address), str(mode))
        self.logger.debug(funcname)
        # Split the metadata into a remote and a local part
        metadata_local = {'metadata':copy.deepcopy(self.statistics['metadata'])}
        metadata_remote = {}
        metadata_remote_query = {}
        # Loop over all keys, check for REMOTE__ and loop over entries
        for metadatakey in self.statistics['metadata']:
            if metadatakey.startswith('REMOTE__'):
                mtmp = metadata_local['metadata'].pop(metadatakey)
                #print('Mtmp',mtmp)
                #print('Mtmp keys', mtmp.keys())
                #print('Mtmp metadatakey', metadatakey)
                raddr_tmp_str = metadatakey.split('REMOTE__')[1]
                raddr_tmp = redvypr_address.RedvyprAddress(raddr_tmp_str)
                remote_address_str = raddr_tmp.get_fullstr()
                metadata_remote[remote_address_str] = {}
                metadata_remote[remote_address_str]['metadata'] = mtmp
                self.logger.debug(funcname + 'Remote query ...')
                metadata_remote_query_tmp = packet_statistic.get_metadata_deviceinfo_all(metadata_remote[remote_address_str], address, mode=mode, publisher_strict=False)
                #print('Remote query',metadata_remote_query)
                metadata_remote_query.update(metadata_remote_query_tmp)

        metadata = packet_statistic.get_metadata(metadata_local, address, mode=mode)
        #print('Metadata (local)', metadata)
        if len(metadata.keys()) > 0:
            metadata_remote_query.update(metadata)

        return metadata_remote_query

    def check_and_fill_config(self):
        """ Fills a config, if essential entries are missing
        """
        try:
            self.custom_config.address
        except:
            self.configaddress = get_ip()


        if(self.custom_config.address == None):
            self.custom_config.address = get_ip()
        elif(self.custom_config.address == ''):
            self.custom_config.address = get_ip()
        elif(self.custom_config.address == '<ip>'):
            self.custom_config.address = get_ip()
        elif(self.custom_config.address == '<IP>'):
            self.custom_config.address = get_ip()
            
        try:
            self.custom_config.port
        except:
            self.custom_config.port = 18196
            
        try:
            self.custom_config.protocol
        except:
            self.custom_config.protocol = 'tcp'
        
        try:    
            self.custom_config.direction
        except:
            self.custom_config.direction = 'publish' # publish/receive
            
        try:
            self.custom_config.datakey
        except: 
            self.custom_config.datakey = 'data' #

        try:
            self.custom_config.serialize
        except:
            self.custom_config.serialize = 'raw' # yaml/str/raw

    def status(self):
        funcname = 'status()'
        # print('Status')
        try:
            status = self.statusqueue.get_nowait()
        except:
            return

        #if type(status) == str:
        #    print('Status', status)
        #else:
        #    print('Statusdict', status)

        return status

    def get_status(self):
        funcname = 'get_status()'
        #print('Status')
        while True:
            try:
                status = self.statusqueue.get_nowait()
            except:
                return

            if type(status) == str:
                #print('Status str',status)
                pass
            else:
                #print('Statusdict',status)
                self.network_status.update(status)
                self.network_status_changed.emit()

        #return statusstr
            

    def __str__(self):
        sstr = 'network device'
        return sstr



class initDeviceWidget(QtWidgets.QWidget):
    #device_start = QtCore.pyqtSignal(Device)
    #device_stop = QtCore.pyqtSignal(Device)
    connect      = QtCore.pyqtSignal(RedvyprDevice) # Signal requesting a connect of the datainqueue with available dataoutqueues of other devices
    def __init__(self,device=None):
        super(QtWidgets.QWidget, self).__init__()
        layout        = QtWidgets.QFormLayout(self)
        self.device   = device
        self.device.check_and_fill_config() # Add standard stuff
        self.inputtabs = QtWidgets.QTabWidget() # Create tabs for different connection types
        self.serialwidget = QtWidgets.QWidget()
        self.label    = QtWidgets.QLabel("Network device")
        self.label.setAlignment(QtCore.Qt.AlignCenter)
        self.label.setStyleSheet("font-weight: bold; font-size: 16")
        self.subbtn = QtWidgets.QPushButton("Subscribe")
        self.subbtn.clicked.connect(self.subscribe_clicked)

        self.startbtn = QtWidgets.QPushButton("Open device")
        self.startbtn.clicked.connect(self.start_clicked)
        self.startbtn.setCheckable(True)

        self.config_widgets = [] # Collecting all config widgets here, makinf it easier to grey them out
        # Address and port
        self.addressline = QtWidgets.QLineEdit()
        self.addressline.setStatusTip('The IP Address, for UDP Broadcast use "<broadcast>", use <IP> for local network IP')
        myip = get_ip()
        addresses = ["<broadcast>", "<IP>", myip ]
        completer = QtWidgets.QCompleter(addresses)
        self.addressline.setCompleter(completer)
        self.portline = QtWidgets.QLineEdit()
        
            
        # Data direction
        self._combo_inout = QtWidgets.QComboBox()
        self._combo_inout.addItem('Publish')        
        self._combo_inout.addItem('Receive')

        
        # Data direction
        self._combo_proto = QtWidgets.QComboBox()
        self._combo_proto.addItem('TCP')        
        self._combo_proto.addItem('UDP')

        # Publish options
        # Create an array of radio buttons
        self._data_pub_all  = QtWidgets.QRadioButton("Redvypr YAML datapacket")
        self._data_pub_all.setStatusTip('Sends/Expects the whole datapacket as a YAML string')


        self._data_pub_dict = QtWidgets.QRadioButton("Dictionary entries")
        self._data_pub_dict.setStatusTip('Sends entries and serialize them as choosen by the serialization box')

        self._data_pub_group = QtWidgets.QButtonGroup()
        self._data_pub_group.addButton(self._data_pub_all)
        self._data_pub_group.addButton(self._data_pub_dict)

        self._data_pub_dict.setChecked(False)
        self._data_pub_all.setChecked(True)
        
        # Serialize
        self._combo_ser = QtWidgets.QComboBox()
        self._combo_ser.addItem('utf-8')
        self._combo_ser.addItem('raw')
        self._combo_ser.addItem('yaml')





        
        
        # Data format to submit
        self.fdataentry  = QtWidgets.QLabel("Data keys publish")        
        self.dataentry   = QtWidgets.QLineEdit()
        self.dataentry.setText('data')


        # The layout of all widgets    
        layout.addRow(self.label)        
        layout.addRow(QtWidgets.QLabel("Address"),self.addressline)  
        layout.addRow(QtWidgets.QLabel("Port"),self.portline)
        layout.addRow(QtWidgets.QLabel("Data direction"),self._combo_inout)
        layout.addRow(QtWidgets.QLabel("Protocol"),self._combo_proto)
        self._serialize_label = QtWidgets.QLabel("Data publishing options")
        layout.addRow(self._serialize_label)
        layout.addRow(self._data_pub_all,self._data_pub_dict)
        layout.addRow(QtWidgets.QLabel("Serialize"),self._combo_ser)
        layout.addRow(self.fdataentry,self.dataentry)
        layout.addRow(self.subbtn)
        layout.addRow(self.startbtn)
        
        self.config_widgets.append(self.addressline)
        self.config_widgets.append(self.portline)
        self.config_widgets.append(self._combo_inout)
        self.config_widgets.append(self._combo_proto)        
        self.config_widgets.append(self._data_pub_all)
        self.config_widgets.append(self._data_pub_dict)
        self.config_widgets.append(self._combo_ser)
        self.config_widgets.append(self.dataentry)

        self.config_to_buttons()
        self.process_options()
        self.connect_signals_options(connect=True)

        self.statustimer = QtCore.QTimer()
        self.statustimer.timeout.connect(self.update_buttons)
        self.statustimer.start(500)

    def subscribe_clicked(self):
        button = self.sender()
        self.connect.emit(self.device)
    def connect_signals_options(self,connect=True):
        if True:
            if(connect):
                self._combo_inout.currentIndexChanged.connect(self.process_options)
                self._combo_proto.currentIndexChanged.connect(self.process_options)
                self._data_pub_all.toggled.connect(self.process_options)
                self._combo_ser.currentIndexChanged.connect(self.process_options)
            else:
                try:
                    self._combo_inout.currentIndexChanged.disconnect()
                    self._combo_proto.currentIndexChanged.disconnect()
                    self._data_pub_all.toggled.disconnect()
                    self._combo_ser.currentIndexChanged.disconnect()
                except:
                    pass

        
    def config_to_buttons(self):
        """ Update the configuration widgets according to the config dictionary in the device module 
        """
        #print('Config to buttons',self.device.config)

        # Disconnect all signals
        self.connect_signals_options(connect=False)
        if (self.device.custom_config.protocol == 'tcp'):
            self._combo_proto.setCurrentIndex(0)
        else:
            self._combo_proto.setCurrentIndex(1)

        if(self.device.custom_config.address is not None):
            self.addressline.setText(self.device.custom_config.address)
        if(self.device.custom_config.port is not None):
            self.portline.setText(str(self.device.custom_config.port))
            
        if(self.device.custom_config.direction == 'publish'):
            self._combo_inout.setCurrentIndex(0)
        else:
            self._combo_inout.setCurrentIndex(1)

        for i in range(self._combo_ser.count()):
            self._combo_ser.setCurrentIndex(i)
            txt = self._combo_ser.currentText()
            #print('txt',i,txt)
            if(txt.lower() == self.device.custom_config.serialize.lower()):
                #print('break')
                break

        if(self.device.custom_config.datakey.lower() == 'all'):
            self._data_pub_all.setChecked(True)
        else:
            self._data_pub_dict.setChecked(True)


        self.dataentry.setText(self.device.custom_config.datakey)
        self.connect_signals_options(connect=True)
            
    def process_options(self):
        """ Reads all options and creates a configuration dictionary
        """
        funcname = __name__ + '.process_options()'
        config   = {}
        logger.debug(funcname)
        config = self.device.custom_config

        config.address   = self.addressline.text()
        config.direction = self._combo_inout.currentText().lower()
        config.protocol  = self._combo_proto.currentText().lower()
        if(self._data_pub_all.isChecked()): # sending/receiving YAML dictionaries
            self.dataentry.setEnabled(False)
            self._combo_ser.setEnabled(False)
            config.serialize = 'yaml'
            config.datakey      = 'all'
        else:
            self.dataentry.setEnabled(True)
            self._combo_ser.setEnabled(True)
            config.serialize = self._combo_ser.currentText().lower()
            config.datakey   = self.dataentry.text()

        # Change the GUI according to send/receive
        if(self._combo_inout.currentText() == 'Publish'):
            self._serialize_label.setText("Data publishing options")
            self.fdataentry.setText("Data key to publish")
        else:
            self._serialize_label.setText("Data receiving options")
            self.fdataentry.setText("Data key to receive")


        # Check for invalid combinations
        if((config.address.lower()) == '<broadcast>' and config.protocol == 'tcp'):
            logger.warning(funcname + ': Broadcast works only with UDP')

        try:
            config.port = int(self.portline.text())
        except:
            logger.warning(funcname + ': Port is not an int')


        self.device.custom_config = config
        logger.debug(funcname + ': config: ' + str(config))


    def thread_status(self,status):
        self.update_buttons(status['threadalive'])
        
    def update_buttons(self):
            """ Updating all buttons depending on the thread status (if its alive, graying out things)
            """

            status = self.device.get_thread_status()
            thread_status = status['thread_running']
            # Running
            if(thread_status):
                self.startbtn.setText('Stop')
                self.startbtn.setChecked(True)
                #for w in self.config_widgets:
                #    w.setEnabled(False)
            # Not running
            else:
                self.startbtn.setText('Start')
                #for w in self.config_widgets:
                #    w.setEnabled(True)
                    
                # Check if an error occured and the startbutton 
                if(self.startbtn.isChecked()):
                    self.startbtn.setChecked(False)
                #self.conbtn.setEnabled(True)

    def start_clicked(self):
        funcname = __name__ + '.start_clicked()'
        button = self.sender()
        if button.isChecked():
            try:
                self.process_options()
            except:
                logger.warning(funcname + ': Invalid settings')
                self.startbtn.setChecked(False)
                return

            # Setting the configuration
            self.device.thread_start()
            button.setText("Starting")
        else:
            logger.debug('Stopping thread now')
            self.device.thread_stop()
            button.setText("Stopping")




class displayDeviceWidget(QtWidgets.QWidget):
    def __init__(self, device=None, tabwidget=None):
        super().__init__()
        self.device = device
        self.layout = QtWidgets.QGridLayout(self)
        self.tabwidget = tabwidget
        #self.statuswidget_network = QtWidgets.QWidget()
        #self.tabwidget.addTab(self.statuswidget_network,'Network devices connected')
        self.statusdictWidget = dictQTreeWidget(self.device.network_status, dataname = 'network status')
        self.device.network_status_changed.connect(self.update_status)
        self.layout.addWidget(self.statusdictWidget)

    def update_status(self):
        """

        """
        status = self.device.network_status
        self.statusdictWidget.reload_data(status)



import copy
import datetime
import redvypr.data_packets as data_packets
from redvypr.widgets.pydanticConfigWidget import pydanticConfigWidget
import numpy
import pydantic
import typing
import sys
import logging
from PyQt5 import QtWidgets, QtCore, QtGui
from redvypr.devices.sensors.calibration.calibration_models import calibration_NTC
from .average_data import average_data
from .sensorWidgets import sensorCoeffWidget

logging.basicConfig(stream=sys.stderr)
logger = logging.getLogger('temperature_array_sensors')
logger.setLevel(logging.DEBUG)

avg_objs_TAR_raw = {}  # Average dataobjects


class parameter_TAR(pydantic.BaseModel):
    NTC_A: typing.Optional[typing.List[calibration_NTC]] = pydantic.Field(default=[])
    name: typing.Optional[typing.List[str]] = pydantic.Field(default=[],description='Name of the parameter')
    pos_x: typing.Optional[typing.List[float]] = pydantic.Field(default=[])
    pos_y: typing.Optional[typing.List[float]] = pydantic.Field(default=[])
    pos_z: typing.Optional[typing.List[float]] = pydantic.Field(default=[])


class sensor_TAR(pydantic.BaseModel):
    description: str = 'Temperature array (TAR) sensor'
    sensor_type: typing.Literal['TAR'] = 'TAR'
    logger_model: str = 'TAR'
    logger_configuration: str = 'TAR'
    parameter: parameter_TAR = parameter_TAR()
    sn: str = pydantic.Field(default='', description='The serial number and/or MAC address of the sensor')
    comment: str = pydantic.Field(default='')
    use_config: bool = pydantic.Field(default=False, description='Use the configuration (True)')
    avg_data: list = pydantic.Field(default=[60, 300], description='List of averaging intervals')

    def init_from_data(self, dataline=None):
        super().__init__()
        if dataline is not None:
            dataparsed = parse_TAR_raw(dataline)
            self.sn = dataparsed['sn']
            n_NTC = len(dataparsed['NTC_A'])
        else:
            return

        # Parse positions, this can be done more elegantly
        dx_positions = float(dataparsed['positions'][1:])
        pos_x = 0
        if n_NTC is not None:
            for i in range(n_NTC):

                if n_NTC > 0:
                    numzeros = int(numpy.floor(numpy.log10(n_NTC))) + 1
                else:
                    numzeros = 1

                #intformat = 'NTC_A{:0' + str(numzeros) + 'd}'
                parameter_name = 'NTC_A[{:d}]'.format(i)
                NTC_name = parameter_name
                ntc_coeff = [1.9560290742146262e-07, -7.749325133113135e-08, 0.00025341112950681254,
                             0.0012350747622505245]
                coeff_comment = 'Standard coefficients, not calibration'
                NTCcal = calibration_NTC(parameter=parameter_name, coeff=ntc_coeff, comment=coeff_comment, sn=self.sn)
                self.parameter.NTC_A.append(NTCcal)
                self.parameter.pos_x.append(pos_x)
                self.parameter.name.append(NTC_name)
                #print('Adding parameter',parameter_name)
                pos_x += dx_positions


def process_TAR_data(dataline, data, device_info, loggerconfig):
    funcname = __name__ + '.process_TAR_data():'
    datapackets_return = []
    try:
        datapacket_TAR = process_TAR_raw(dataline, data, device_info)
        datapacket_TAR['datatype'] = 'raw'
    except:
        logger.info(' Could not decode {:s}'.format(str(dataline)), exc_info=True)
        datapacket_TAR = None

    if datapacket_TAR is not None:

        datapackets_return.append(datapacket_TAR)
        # Convert to temperature
        devicename = 'TAR_SI_' + datapacket_TAR['sn']
        datapacket_TAR_T = data_packets.create_datadict(device=devicename, tu=datapacket_TAR['_redvypr']['t'],
                                                        hostinfo=device_info['hostinfo'])
        datapacket_TAR_T['type'] = datapacket_TAR['type']
        datapacket_TAR_T['np'] = datapacket_TAR['np']
        datapacket_TAR_T['sn'] = datapacket_TAR['sn']
        datapacket_TAR_T['t'] = datapacket_TAR['t']
        datapacket_TAR_T['datatype'] = 'converted'
        datapacket_TAR_T['sensortype'] = datapacket_TAR['sensortype']
        datapacket_TAR_T['NTC_A'] = []
        j = 0
        for i, R in enumerate(datapacket_TAR['NTC_A']):
            j += 1
            #print('COEFF',loggerconfig.parameter.NTC_A)
            poly = loggerconfig.parameter.NTC_A[i].coeff
            #poly = [1.9560290742146262e-07, -7.749325133113135e-08, 0.00025341112950681254, 0.0012350747622505245]
            Toff = 273.15
            # print('Poly', poly)
            data_tmp = numpy.polyval(poly, numpy.log(R))
            # print('data_tmp', data_tmp)
            data_tmp = 1 / data_tmp - Toff
            dataSI = data_tmp
            # print('data_SI', dataSI,type(dataSI))
            # datapacket[parameter] = float(dataSI)
            # And finally the data
            datapacket_TAR_T['NTC_A'].append(float(dataSI))

        datapackets_return.append(datapacket_TAR_T)
        #print('Datapacket TAR', datapacket_TAR_T)
        # dataqueue.put(datapacket_TAR)
        # Do averaging.
        sn = datapacket_TAR['sn']
        if False:
            try:
                datapacket_TAR_avg = avg_objs_TAR_raw[sn].average_data(data)
            except:
                parameters = ['t', 'NTC_A']
                avg_objs_TAR_raw[sn] = average_data(datapacket_TAR, parameters,
                                            loggerconfig)
                datapacket_TAR_avg = avg_objs_TAR_raw[sn].average_data(datapacket_TAR)
                logger.debug(funcname + ' could not average:', exc_info=True)

            if datapacket_TAR_avg is not None:
                datapackets_return.append(datapacket_TAR_avg)

    return datapackets_return


def process_TAR_raw(dataline, data, device_info):
    datapacket = parse_TAR_raw(dataline)
    if datapacket is not None:
        macstr = datapacket['sn']
        datapacket['t'] = data['t']
        devicename = 'TAR_raw_' + macstr
        datapacket_redvypr = data_packets.create_datadict(device=devicename, tu=data['_redvypr']['t'],
                                                          hostinfo=device_info['hostinfo'])
        datapacket_redvypr.update(datapacket)

        return datapacket_redvypr


# Temperature array
def parse_TAR_raw(data, sensortype='TAR'):
    """
    Parses temperature array
    $FC0FE7FFFE155D8C,TAR,B2,36533.125000,83117,3498.870,3499.174,3529.739,3490.359,3462.923,3467.226,3480.077,3443.092,3523.642,3525.567,3509.492,3561.330,3565.615,3486.693,3588.670,3539.169,3575.104,3523.946,3496.343,3480.160,3531.045,3501.624,3497.010,3557.235,3479.952,3458.297,3523.052,3487.223,3571.087,3525.740,3580.928,3534.818
    """
    #print('Got data', data)
    datasplit = data.split(',')
    macstr = datasplit[0][1:17]
    packettype = datasplit[1]  # HF
    datapacket = {}
    datapacket['type'] = packettype
    datapacket['sn'] = macstr
    datapacket['sensortype'] = sensortype
    valid_package = False
    numsp = len(datasplit)
    if numsp > 3:
        ts = float(datasplit[3])  # Sampling time
        np = int(datasplit[4])  # Number of packet
        datapacket['ts'] = ts
        datapacket['np'] = np
        datapacket['positions'] = datasplit[2]
        ntc_all = []
        j = 0
        for i in range(5, numsp):
            j += 1
            data = float(datasplit[i])
            ntc_all.append(data)

        datapacket['NTC_A'] = ntc_all

        valid_package = True

    if valid_package:
        return datapacket
    else:
        return None


class TARWidget(QtWidgets.QWidget):
    """
    Widget to display temperature array sensor data
    """
    def __init__(self,*args, sn=None, redvypr_device=None):
        funcname = __name__ + '__init__()'
        logger.debug(funcname)
        super(QtWidgets.QWidget, self).__init__(*args)

        self.device = redvypr_device
        self.sn = sn
        self.configuration = None
        if self.device is not None:
            try:
                self.configuration = self.device.custom_config.sensorconfigurations[sn]
            except Exception as e:
                logger.exception(e)

        self.pos_x = self.configuration.parameter.pos_x
        self.parameter = ['NTC_A']
        self.parameter_unit = []
        self.parameter_unitconv = []
        #print('Adding parameter')
        self.rawconvtabs = QtWidgets.QTabWidget()
        self.rawwidget = QtWidgets.QWidget()
        self.layout_all = QtWidgets.QGridLayout(self.rawwidget)
        self.convwidget = QtWidgets.QWidget()
        self.layout_conv = QtWidgets.QGridLayout(self.convwidget)
        self.rawconvtabs.addTab(self.rawwidget,'Rawdata')
        self.rawconvtabs.addTab(self.convwidget, 'Converted data')

        self.layout = QtWidgets.QGridLayout(self)
        self.layout.addWidget(self.rawconvtabs,0,0)

        # Add a table for the raw data
        self.datatable_all = QtWidgets.QTableWidget()
        self.datatable_all_columnheader = ['Name','Pos']
        for i, p in enumerate(self.parameter):
            self.datatable_all_columnheader.append(p)
            self.datatable_all_columnheader.append(p)

        nNTC = len(self.configuration.parameter.NTC_A)
        nRows = nNTC + 3 # Add unit, time and numpacket
        self.nRow_data = 3
        self.datatable_all.setRowCount(nRows)

        self.datatable_all.setColumnCount(len(self.datatable_all_columnheader))
        self.datatable_all.setHorizontalHeaderLabels(self.datatable_all_columnheader)
        self.datatable_all.horizontalHeader().setSectionResizeMode(QtWidgets.QHeaderView.Stretch)
        self.datatable_all.verticalHeader().setSectionResizeMode(QtWidgets.QHeaderView.Stretch)
        self.layout_all.addWidget(self.datatable_all,0,1)

    def update_data(self, data):
        """
        Updating the local data with datapacket
        """
        funcname = __name__ + '.update_data()'
        logger.debug(funcname)
        devicename = data['_redvypr']['device']
        #print('Devicename',devicename)
        dkey = self.parameter[0]
        NTC_A = data[dkey]
        self.nRow_time = 1
        self.nRow_np = 2
        self.nRow_unit = 0
        self.nCol_pos = 1
        if 'raw' in devicename:
            nCol_data = 2
            unit = 'Ohm'
        if 'SI' in devicename:
            nCol_data = 3
            unit = 'degC'

        # Add unit, time, numpackage
        item = QtWidgets.QTableWidgetItem('Unit')
        self.datatable_all.setItem(self.nRow_unit, 0, item)
        item = QtWidgets.QTableWidgetItem('Time')
        self.datatable_all.setItem(self.nRow_time, 0, item)
        item = QtWidgets.QTableWidgetItem('Numsample')
        self.datatable_all.setItem(self.nRow_np, 0, item)

        item = QtWidgets.QTableWidgetItem(unit)
        self.datatable_all.setItem(self.nRow_unit, nCol_data, item)
        td = datetime.datetime.fromtimestamp(data['t'])
        tstr = td.strftime('%d %b %Y %H:%M:%S.%f')
        item = QtWidgets.QTableWidgetItem(tstr)
        self.datatable_all.setItem(self.nRow_time, nCol_data, item)
        npstr = str(data['np'])
        item = QtWidgets.QTableWidgetItem(npstr)
        self.datatable_all.setItem(self.nRow_np, nCol_data, item)

        for i,NTC in enumerate(NTC_A):
            # Name
            namestr = "NTC{:02d}".format(i+1)
            item = QtWidgets.QTableWidgetItem(namestr)
            self.datatable_all.setItem(i+self.nRow_data,0,item)
            # Position
            pos = self.pos_x[i]
            item = QtWidgets.QTableWidgetItem(str(pos))
            self.datatable_all.setItem(i + self.nRow_data, self.nCol_pos, item)
            # Data
            dstr = "{:.3f}".format(NTC)
            item = QtWidgets.QTableWidgetItem(dstr)
            self.datatable_all.setItem(i+self.nRow_data,nCol_data,item)

        # Fill the datatables
        self.datatable_all.resizeColumnsToContents()


#
#
# Here the widgets start
#
#
class TARWidget_config(QtWidgets.QWidget):
    """
    Widget to configure a temperature array sensor
    """

    def __init__(self, *args, sn=None, redvypr_device=None):
        funcname = __name__ + '__init__()'
        super(QtWidgets.QWidget, self).__init__(*args)
        logger.debug(funcname)
        self.device = redvypr_device
        self.nosensorstr = ''
        self.sn = sn
        # Try to find a configuration for the logger
        try:
            conf = self.device.custom_config.sensorconfigurations[self.sn]
            logger.debug(funcname + ' Found a configuration')
        except:
            logger.warning(funcname + ' Did not find a configuration')
            return

        self.config = conf
        conflocal = copy.deepcopy(conf)
        self.configWidget = pydanticConfigWidget(self.config, exclude=['parameter'])

        self.parameterWidget = QtWidgets.QWidget(self)
        self.fill_parameter_widgets()

        self.layout = QtWidgets.QGridLayout(self)
        self.layout.addWidget(self.configWidget,0,0)
        self.layout.addWidget(self.parameterWidget,0,1)

    def __assign_calibrations__(self):
        funcname = __name__ + '__assign_calibrations():'
        print(funcname)
        print('calibrations', self.device.custom_config.calibrations)
        cals = self.device.custom_config.calibrations
        match = ['parameter','sn']
        for i, para in enumerate(self.config.parameter.NTC_A):
            print('Searching for calibration for parameter',i)
            print('Para',i,para)
            cal_match = []
            cal_match_date = []
            for cal in cals:
                match_all = True
                for m in match:
                    mcal = getattr(cal, m)
                    mpara = getattr(para, m)
                    if mcal != mpara:
                        match_all = False

                if match_all:
                    print('Found matching parameter')
                    print('Cal',cal)
                    cal_match.append(cal)
                    td = datetime.datetime.strptime(cal.date,'%Y-%m-%d %H:%M:%S.%f')
                    cal_match_date.append(td)

            if len(cal_match) > 0:
                imin = numpy.argmin(cal_match)
                print('Assigning matching parameter')
                self.config.parameter.NTC_A[i] = cal_match[imin]

        self.__fill_calibration_table__()

    def __create_calibration_widget__(self):
        self.__calbutton_clicked__ = self.sender()
        calibrations = redvypr_device = self.device.custom_config.calibrations
        self.__calwidget__ = sensorCoeffWidget(calibrations=calibrations, redvypr_device=self.device)
        self.applyCoeffButton = QtWidgets.QPushButton('Apply')
        self.applyCoeffButton.clicked.connect(self.applyCalibration_clicked)
        self.cancelCoeffButton = QtWidgets.QPushButton('Cancel')
        self.cancelCoeffButton.clicked.connect(self.applyCalibration_clicked)
        self.__calwidget__.sensorCoeffWidget_layout.addWidget(self.applyCoeffButton, 2, 0)
        self.__calwidget__.sensorCoeffWidget_layout.addWidget(self.cancelCoeffButton, 2, 1)
        self.__calwidget__.show()

    def applyCalibration_clicked(self):
        if self.sender() == self.cancelCoeffButton:
            self.sensorCoeffWidget.close()
        else:
            print('Apply')
            user_role = 10
            item = self.__calwidget__.sensorCoeffWidget_list.currentItem()
            if item is not None:
                role = QtCore.Qt.UserRole + user_role
                print('fds', self.__calwidget__.sensorCoeffWidget_list.currentRow(), item.data(role))
                cal = item.data(role)
                print('Cal',cal)
                self.__cal_apply__ = cal # Save the calibration
                #Update the calibration
                iupdate = self.__calbutton_clicked__.__para_index__
                print('Iupdate', iupdate)
                self.config.parameter.NTC_A[iupdate] = cal
                self.__calwidget__.close()
                self.__fill_calibration_table__()


    def __fill_calibration_table__(self):
        funcname = __name__ + '__fill_calibration_table__():'
        logger.debug(funcname)
        self.parameterTable.clear()
        nRows = len(self.config.parameter.NTC_A)
        nCols = 5
        self.parameterTable.setRowCount(nRows)
        self.parameterTable.setColumnCount(nCols)
        self.parameterTable.setHorizontalHeaderLabels(['Sensor Parameter','Cal SN','Cal Parameter','Cal Date','Cal Comment'])
        #print('Config parameter',self.config.parameter)
        for i,para in enumerate(self.config.parameter.NTC_A):
            #print('Para',para)
            name = self.config.parameter.name[i]
            but = QtWidgets.QPushButton(name)

            #but.clicked.connect(self.parent().show_coeffwidget_apply)
            but.clicked.connect(self.__create_calibration_widget__)
            but.__para__ = para
            but.__para_index__ = i
            self.parameterTable.setCellWidget(i,0,but)
            # SN
            item = QtWidgets.QTableWidgetItem(para.sn)
            self.parameterTable.setItem(i, 1, item)
            # Parameter
            item = QtWidgets.QTableWidgetItem(para.parameter)
            self.parameterTable.setItem(i, 2, item)
            # Date
            item = QtWidgets.QTableWidgetItem(para.date)
            self.parameterTable.setItem(i, 3, item)
            # Comment
            item = QtWidgets.QTableWidgetItem(para.comment)
            self.parameterTable.setItem(i, 4, item)

        self.parameterTable.resizeColumnsToContents()
    def fill_parameter_widgets(self):
        funcname = __name__ +'.fill_parameter_widgets():'
        self.parameterLayout = QtWidgets.QGridLayout(self.parameterWidget)
        self.parameterAuto = QtWidgets.QPushButton('Autofill calibrations')
        self.parameterAuto.clicked.connect(self.__assign_calibrations__)
        self.parameterTable = QtWidgets.QTableWidget()

        self.auto_check_sn = QtWidgets.QCheckBox('SN')
        self.auto_check_sn.setEnabled(False)
        self.auto_check_sn_edit = QtWidgets.QLineEdit()
        self.auto_check_sn_edit.setText(self.sn)
        self.auto_check_sn_edit.setEnabled(False)
        self.parameterLayout.addWidget(self.parameterAuto, 0, 0, 1 , 2)
        self.parameterLayout.addWidget(self.auto_check_sn, 1, 0)
        self.parameterLayout.addWidget(self.auto_check_sn_edit, 1, 1)
        self.parameterLayout.addWidget(self.parameterTable, 2, 0, 1, 2)
        self.__fill_calibration_table__()



from PyQt5 import QtWidgets, QtGui, QtCore
import logging
import sys
from redvypr.redvypr_address import RedvyprAddress


logging.basicConfig(stream=sys.stderr)
logger = logging.getLogger('RedvyprSubsribeWidget')
logger.setLevel(logging.DEBUG)

class LineEditFocus(QtWidgets.QLineEdit):
    focusInSignal = QtCore.pyqtSignal()
    focusOutSignal = QtCore.pyqtSignal()
    def focusInEvent(self, event):
        # do custom stuff
        self.focusInSignal.emit()
        super(LineEditFocus, self).focusInEvent(event)

    def focusOutEvent(self, event):
        # do custom stuff
        self.focusOutSignal.emit()
        super(LineEditFocus, self).focusOutEvent(event)
class redvyprSubscribeWidget(QtWidgets.QWidget):
    """ Widget that lets the user add/modify/remove subscriptions of a device

    """

    def __init__(self, redvypr=None, device=None, show_devices = False):
        """

        Args:
            redvypr:
            device: The device the user can change the
            show_devices: lets the user choose between different devices, otherwise only the subscriptions of device=device can be changed.
        """
        super(redvyprSubscribeWidget, self).__init__()
        self.show_devices = show_devices
        self.redvypr = redvypr
        self.devices = self.redvypr.devices
        self.redvypr.devices_connected.connect(self.__devices_connected__)
        self.redvypr.devices_disconnected.connect(self.__devices_connected__)
        self.redvypr.device_status_changed_signal.connect(self.__devices_connected__)
        if (len(self.devices) > 0):
            if (device == None):  # Take the first one
                device = self.devices[0]['device']

        # Set icon
        #self.setWindowIcon(QtGui.QIcon(_icon_file))
        self.setWindowTitle("redvypr subscribe")
        layout = QtWidgets.QGridLayout(self)

        font = QtGui.QFont('Arial', 20)
        font.setBold(True)
        lab = QtWidgets.QLabel('Device subscriptions')
        lab.setFont(font)
        lab.setAlignment(QtCore.Qt.AlignCenter)
        self.lab = lab

        self.device_label = QtWidgets.QLabel('Device')
        self.device_label.setAlignment(QtCore.Qt.AlignCenter)
        self.dataprovider_label = QtWidgets.QLabel('Data providing devices')
        self.dataprovider_label.setAlignment(QtCore.Qt.AlignCenter)
        self.subscribe_label = QtWidgets.QLabel('Subscriptions')
        self.subscribe_label.setAlignment(QtCore.Qt.AlignCenter)

        self.devices_listPublisher = QtWidgets.QTreeWidget()  # All data publishing devices
        self.devices_listPublisher.setColumnCount(2)
        self.devices_listPublisher.setHeaderHidden(True)
        self.devices_listPublisher.currentItemChanged.connect(self.__update_device_choice__)

        self.devices_listDevices = QtWidgets.QListWidget()  # The devices a connection is to be defined
        self.devices_listDevices.itemClicked.connect(self.itemcon_clicked)
        self.devices_listDevices.itemDoubleClicked.connect(self.itemcon_dclicked)

        self.devices_listallsub = QtWidgets.QListWidget()  # The subscriptions of the device
        self.devices_listallsub.itemClicked.connect(self.__itemsubscribed_clicked__)

        self.subscribe_edit = LineEditFocus()
        self.subscribe_edit.focusInSignal.connect(self.__focus_in__)
        self.subscribe_edit.focusOutSignal.connect(self.__focus_out__)
        self.subscribe_edit.redvypr_address = RedvyprAddress('*')

        self.__commitbtn = QtWidgets.QPushButton('Subscribe')
        self.__commitbtn.clicked.connect(self.commit_clicked)
        self.__commitbtn.setEnabled(False)

        self.__subscribeAllBtn = QtWidgets.QPushButton('Subscribe all (*)')
        self.__subscribeAllBtn.clicked.connect(self.subscribeAll_clicked)

        self.__closeBtn = QtWidgets.QPushButton('Close')
        self.__closeBtn.clicked.connect(self.close_clicked)
        #self.__commitbtn.setEnabled(False)
        # Combo with formats
        self.__formatLabel = QtWidgets.QLabel('Address format')
        addr_formats = RedvyprAddress().get_common_address_formats()
        self.__formatCombo = QtWidgets.QComboBox()
        self.__formatCombo.currentIndexChanged.connect(self.__subscribe_editChanged__)
        for f in addr_formats:
            self.__formatCombo.addItem(f)

        self.__formatCombo.setCurrentIndex(0)
        #layout.addWidget(lab, 0, 1,1,3)
        layout.addWidget(lab, 0, 0,1,4)

        if self.show_devices:
            layout.addWidget(self.device_label, 1, 0)
            layout.addWidget(self.devices_listDevices, 2, 0)


        layout.addWidget(self.subscribe_label, 1, 0 , 1, 2)
        layout.addWidget(self.dataprovider_label, 1, 2, 1, 2)
        layout.addWidget(self.devices_listallsub, 2, 0 ,1,2)
        layout.addWidget(self.devices_listPublisher, 2, 2, 1, 2)
        layout.addWidget(self.subscribe_edit, 3, 0, 1, 2)
        layout.addWidget(self.__formatLabel, 3, 2, 1, 1)
        layout.addWidget(self.__formatCombo, 3, 3, 1, 1)
        layout.addWidget(self.__commitbtn,4,0,1,-1)
        layout.addWidget(self.__subscribeAllBtn, 5, 0, 1, -1)
        layout.addWidget(self.__closeBtn, 6, 0, 1, -1)

        if (len(self.devices) > 0):
            self.update_list(device)

    def close_clicked(self):
        self.close()

    def __itemsubscribed_clicked__(self,item):
        self.__commitbtn.setText('Remove')
        self.__commitbtn.setEnabled(True)
        self.__commitbtn.__status__ = 'remove'
        self.__commitbtn.redvypr_addr_remove = item.redvypr_addr

    def __focus_in__(self):
        self.__commitbtn.setText('Subscribe')
        self.__commitbtn.__status__ = 'add'
        self.__commitbtn.setEnabled(True)

    def __focus_out__(self):
        pass

    def __devices_connected__(self, dev1=None, dev2=None):
        #print('Devices have been connected',dev1,dev2)
        self.update_list(self.device)

    def __subscribe_editChanged__(self):
        addr_format = self.__formatCombo.currentText()
        devstr = self.subscribe_edit.redvypr_address.get_str(addr_format)
        self.subscribe_edit.setText(devstr)

    def __update_device_choice__(self, newitem, olditem):
        """
        A device was clicked, update all buttons
        Args:
            item:

        Returns:

        """
        #addr_format = self.__formatCombo.currentText()
        if newitem is not None:
            #devstr = newitem.redvypr_address.get_str(addr_format)

            try:
                subscribed = newitem.subscribed
            except:
                subscribed = False

            self.__commitbtn.setText('Subscribe')
            self.__commitbtn.__status__ = 'add'
            self.__commitbtn.setEnabled(True)
            #self.__commitbtn.redvypr_addr_remove = devstr
            self.subscribe_edit.redvypr_address = newitem.redvypr_address#.RedvyprAddress
            self.__subscribe_editChanged__()

            #if(subscribed):
            #    self.__commitbtn.setText('Unsubscribe')
            #    self.__commitbtn.__status__ = 'remove'
            #    self.__commitbtn.setEnabled(True)
            #    self.__commitbtn.redvypr_addr_remove = devstr
            #else:
            #    self.subscribe_edit.setText(devstr)
            #    #print(devstr)
            #    #print('Item',newitem.text(0))
            #    self.__commitbtn.setText('Subscribe')
            #    self.__commitbtn.__status__ = 'add'
            #    self.__commitbtn.setEnabled(True)
        else:
            self.__commitbtn.setEnabled(False)

    def update_list(self, device):
        """ Update the list
        """

        funcname = __name__ + '.update_list()'
        try:
            devname = device.name
        except:
            devname = 'NA'
        logger.debug(funcname + ':update_list for device: {:s}, name {:s}'.format(str(device),devname))
        self.devices_listPublisher.clear()
        self.devices_listallsub.clear()
        self.devices_listDevices.clear()
        self.device = device

        if (len(self.devices) > 0):
            root = self.devices_listPublisher.invisibleRootItem()
            # self.devices_listDevices.addItem(str(device))
            data_provider_all = self.redvypr.get_device_objects(publishes=True, subscribes=False)
            font1 = QtGui.QFont('Arial')
            font1.setBold(True)
            font0 = QtGui.QFont('Arial')

            # Fill the qtreewidget
            #print('data provider',data_provider_all)
            if (data_provider_all is not None):
                for dev in data_provider_all:
                    if dev == self.device:
                        continue

                    # Check if the device is already subscribed
                    subscribed = False
                    #print('dev',dev.name,dev.redvypr.hostinfo)
                    for a in self.device.subscribed_addresses:
                        subscribed = a in dev.address
                        if subscribed:
                            break

                    itm = QtWidgets.QTreeWidgetItem([dev.name, ''])
                    itm.device = dev
                    itm.redvypr_address = dev.address
                    if subscribed:
                        status = 'subscribed'
                        itm.setFont(0, font1)
                        itm.subscribed = True
                    else:
                        itm.setFont(0, font0)
                        itm.subscribed = False

                    root.addChild(itm)
                    # Add all data_devices
                    if True:
                        devs_forwarded = dev.get_device_info()
                        for devaddress in devs_forwarded.keys():
                            devaddress_redvypr = RedvyprAddress(devaddress)
                            subscribed = False
                            for a in self.device.subscribed_addresses:
                                subscribed = a in devaddress_redvypr
                                if subscribed:
                                    break

                            devaddress_str = devaddress_redvypr.get_str('/a/h/p/i/d/')
                            itmf = QtWidgets.QTreeWidgetItem([devaddress_str, ''])
                            #itmf.setData(0, 0, devaddress_redvypr)
                            itmf.device = dev
                            itmf.redvypr_address = devaddress_redvypr
                            itmf.address_forwarded = devaddress
                            if(subscribed):
                                itmf.setFont(0, font1)
                                itmf.subscribed = True
                            else:
                                itmf.setFont(0, font0)
                                itmf.subscribed = False

                            itm.addChild(itmf)

            self.devices_listPublisher.expandAll()
            self.devices_listPublisher.resizeColumnToContents(0)

            # Fill list of devices subscribing
            devitm = None
            if True:
                # connecting devices
                for s in self.devices:
                    sen = s['device']
                    itm = QtWidgets.QListWidgetItem(sen.name)
                    itm.device = sen
                    self.devices_listDevices.addItem(itm)
                    if (sen == device):
                        devitm = itm

                if(devitm is not None):
                    self.devices_listDevices.setCurrentItem(devitm)
                    self.lab.setText('Subscriptions for\n ' + str(sen.name))

            # Fill the subscribed list
            if True:
                # connecting devices
                for s in self.device.subscribed_addresses:
                    sstr = s.address_str
                    litm = QtWidgets.QListWidgetItem(sstr)
                    litm.redvypr_addr = s
                    self.devices_listallsub.addItem(litm)

    # End update_list()
    def commit_clicked(self):
        """ Apply changes to subscribe/unsubscribe
        """
        funcname = 'commit_clicked'
        logger.debug(funcname)

        if (self.device is not None):
            if self.__commitbtn.__status__ == 'add':
                address_add = str(self.subscribe_edit.text())
                logger.debug('Adding? {}'.format(address_add))
                if (len(address_add) > 0):
                    logger.debug('Adding {}'.format(address_add))
                    self.device.subscribe_address(address_add)
                    self.update_list(self.device)
                else:
                    logger.debug('Nothing to add')
            elif self.__commitbtn.__status__ == 'remove':
                raddr = self.__commitbtn.redvypr_addr_remove
                self.device.unsubscribe_address(raddr)
                self.update_list(self.device)


        getSelected = self.devices_listPublisher.selectedItems()
        if getSelected:
            itm = getSelected[0]
            try:
                device = itm.device
                devicename = device.name
            except:
                device = None
                devicename = ''

            # Get subscriber
            subscriber_item  = self.devices_listDevices.currentItem()
            subscriber = subscriber_item.device
            try:
                address_forwarded = itm.address_forwarded
            except:
                address_forwarded = None

    def subscribeAll_clicked(self):
        logger.debug('Subscribe all')
        self.device.subscribe_address('*')
        self.update_list(self.device)
        self.close()

    def disconnect_clicked(self):
        logger.debug('Disconnect')

    def itemcon_clicked(self, item):
        # Update the connection list
        self.update_list(item.device)

    def itemcon_dclicked(self, item):
        if (item.isSelected()):
            item.setSelected(False)
